c.......................................................................
c
c
c globs.inc:   Global variables to the distributed AMR code
c
c
c.......................................................................

      !
      ! Maximum number of Levels possible
      !
      integer     maxlev
      parameter ( maxlev         =  25        )

      !
      ! Various other globals:
      !
      ! epsilon for K-O dissipation (e.g. 0.05)
      real(kind=8)  epsdis,
     *        lambda,           ! CFL ratio dt/dx (e.g. 0.3)
     *        minx0, miny0, minz0,! User selected bounds on coarse grid
     *        maxx0, maxy0, maxz0,
     *        ref_level(0:maxlev-1), ! refinement factor for level
     *        ell_epsilon       ! Threshold for residual in elliptic

      integer trace_level,     ! determines how much output tracing
                               !   0      minimal time step output
                               ! ( 0,10]  output from main() only
     *        steps,           ! loop counter for coarse grid evolution
     *        refine_factor,   ! ratio dx_1/dx_2 of grid resolutions
                               ! must be even
     *        weno_interp,      ! Determine if using weno or not                         
     *        refine_period,   ! number of time steps between refinement
     *        simpleFMR,       ! flag to use simple volume based FMR
     *        clusterDD,       ! controls whether domain decomposition 
     *        clusterstyle,    ! restricts clustering
     *        diss_afterinj,   ! add dissipation after injection?(if!=0)
     *        Levelp(0:maxlev-1),! pointer to 'first' grid @ given level
                               ! negative if no grids @ that level
     *        out_period(0:maxlev-1),! output period for each level
     *        lev_count(0:maxlev-1), ! step counter for various levels
     *        output_style,  ! output by: level(1), together(2), both(3)
     *        output_dim,    ! dimension of output
     *        linearbounds,  ! use linear interp if > 0
     *        shadow,        ! shadow hierarchy
     *        allowedl,      ! maximum # of levels allowed by user
     *        num_evol_iters,! # of iterations for an evol step
     *        bound_width,   ! width of overlap region for DD
     *        num_injfields, ! number of fields which get injected
     *        num_anafields, ! number of fields requiring analysis
     *        chkpt_period,  ! how often to dump state for checkpoint
     *        chkpt_readstate,! how often to dump state for checkpoint
     *        chkpt_fnum,    ! number/index to chkpt_fname array
     *        findhorizon,   ! how to find a horizon (or use static)
     *        mask_period,   ! how often to find a horizon
     *        mask_usemin,   ! use min of lapse to determine mask guess?
     *        mask_minfield  ! number of field to take minimum of

       character(4) tag      ! prefix for output sdf files
       character(128) chkpt_fname(2) ! name of file for saving state

      !
      ! Clustering parameters:
      !
       real(kind=8)  ethreshold,! error threshold to flag for rfnmnt
     *         minefficiency ! minimum efficiency for new refined grids
       integer buffer,  ! number of coarse grid pts around flagged
     *         mindim,  ! smallest # of coarse grid pts for subgrid
     *         window,  ! # of coarse grid pts from bdy in which
                                     !   clusterer looks to slice box
     *         ghostwidth,        ! Width of region set  by parent
     *         update_scheme,     ! choose RK3 or ICN
     *         amrbound_prepost,  ! enforce subgrd bnds before evolve?
     *         amrbound_timealign,! enforce subgrd only timealigned?
     *         elliptic_solve,    ! ==0 no elliptic solve, >0 then do so
     *         nvcycle,           ! # of vcycles to carry-out
     *         preswp,            ! # of pre-sweeps  of relax. in vcycle
     *         pstswp,            ! # of post-sweeps of relax. in vcycle
     *         maxsweeps,         ! max # of relaxation sweeps for solve
     *         clusterreadwrite   ! Flag:
                                  !   0 --- normal clustering
                                  !   1 --- read  in  hierarchy information
                                  !   2 --- write out hierarchy information

       real(kind=8) total_work! diagnostic measure of integrated "work"
                              !   done relative to coarse grid.

      !
      ! Store maximum value of chi
      !   to detect singularity formation:
      !
      real(kind=8)  maxchi,   ! maximum (absolute) value of chi
     *        maxchi_t        ! time at which max is achieved
      integer maxchi_i, maxchi_j, maxchi_k, ! gridpoint of maximum
     *        maxchi_l,       ! level of maximum
     *        maxchi_reach    ! =0 if not reached collapsed time

      !
      ! Interrupt handling:
      !
      integer interrupt_set,  !  set >0 if interrupt detected
     *        interrupt_rc    !  return code for signal() call

      !
      ! File unit number for output of runtime statistics
      !
      integer fileout

      !
      ! Bounding boxes for refinement: (not in a common block)
      !

      integer     maxnumchildren ! total # of children for any grid
      parameter ( maxnumchildren = 512        )

      !
      ! Flagging constants:
      !
      !     ...points     to be refined
      real(kind=8)   FLAG_REFINE
      parameter    ( FLAG_REFINE   =  1.d0 )
      !     ...points NOT to be refined
      real(kind=8)   FLAG_NOREFINE
      parameter    ( FLAG_NOREFINE =  0.d0 )
      !     ...points which do not exist on the level
      real(kind=8)   FLAG_DISALLOW
      parameter    ( FLAG_DISALLOW = -1.d0 )

      common  / myglobals_doubles /
     *        ref_level,     epsdis,        lambda,
     *        minx0,         miny0,         minz0,
     *        maxx0,         maxy0,         maxz0,
     *        ethreshold,    minefficiency, total_work,
     *        maxchi,        maxchi_t,      ell_epsilon


      common  / myglobals_ints /
     *        out_period,    lev_count,      Levelp,
     *        steps,         ghostwidth,     update_scheme,
     *        amrbound_prepost,              amrbound_timealign,
     *        elliptic_solve, clusterreadwrite,
     *        nvcycle, preswp, pstswp, maxsweeps,
     *        window,        buffer,         mindim,
     *        trace_level,   refine_factor,  linearbounds, shadow,
     *        allowedl,num_evol_iters, bound_width, num_injfields,
     *        num_anafields,
     *        chkpt_period, chkpt_readstate, chkpt_fnum,
     *        findhorizon,  mask_period, mask_usemin, mask_minfield,
     *        refine_period, clusterDD, clusterstyle, simpleFMR,
     *        diss_afterinj,  output_style, output_dim,
     *        maxchi_i,      maxchi_j,       maxchi_k,
     *        maxchi_l,      maxchi_reach,   
     *        interrupt_set,  fileout

      common  / myglobl_strings / tag, chkpt_fname
