#include "cctk.h"

!--------------- R E A D   M E ------------------------------------------
!
! Note as of Fri Mar 14 10:17:45 CST 2003
!
! Currently this routine is very simple minded.  It does the following:
!   (a) Assigns pointers to u, v, w, as well as the derivatives,
!   (b) Defines the parameter array
!   (c) Calls auxvars to define v
!   (d) Calls hyperderivs to calculate dxu, dyu, dzu, dxv, dyv, dzv
!   (e) Calls a general analysis routine that should be in the equation thorn,
!       which is not a point-wise routine.       
!
! The analysis functions are presumably stored in w, but no assumptions
! at this level are made.  NOTE: Point-wise derivatives are NOT supported.
! In the future a point-wise analysis routine probably should be written.
!
!------------------------------------------------------------------------


  subroutine HyperAnalysis(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
     urk2, urk3, par)
  use params
  use GF
  use UTILEQS
  use MOD_ANALYSIS
  use mod_mask
  use hypercoords
  use HYPER_AUXVARS
  use HYPER_DERIVS


  implicit none
  
  type(gridfunction), dimension(NU)    :: u2, u0, dxu, dyu, dzu
  type(gridfunction), dimension(NU)    :: urk1, urk2, urk3

  type(gridfunction), dimension(NV)    :: v, dxv, dyv, dzv
  type(gridfunction), dimension(NW)    :: w
  CCTK_REAL, dimension(NPAR)           :: par
  
  CCTK_REAL, dimension(NU)             :: u_pt,u0_pt,dxu_pt,dyu_pt,dzu_pt
  CCTK_REAL, dimension(NV)             :: v_pt,dxv_pt,dyv_pt,dzv_pt
  CCTK_REAL, dimension(NW)             :: w_pt

  CCTK_INT  :: ii, jj ,kk, m, nx, ny, nz, nt, istat, err, use_mask, imask_pt, order   
  CCTK_INT  :: ilo,ihi,jlo,jhi,klo,khi,rank,ierr, ng
   
  CCTK_INT, dimension(:,:,:), allocatable :: imask
  logical, parameter                   :: ltrace = .false.

  if (ltrace) then
    write(0,*)' >>> HyperAnalysis'
    write(0,*)' >>> p_point_wise_derivatives = ',P_POINT_WISE_DERIVATIVES
    write(0,*)' >>> p_point_wise_analysis = ',nint(par(P_POINT_WISE_ANALYSIS))
  end if
 
  nx = nint(par(P_NX))
  ny = nint(par(P_NY))
  nz = nint(par(P_NZ))
  ng = par(P_BOUND_WIDTH)  
  use_mask = nint(par(P_USE_MASK))
  order = nint(par(P_DERIV_ORDER))

  
  !define coordinates if not using the default cartesian computational coords.
  if (P_ENABLE_ALT_COORDS==1) then
    call def_alt_coords(w, par)
  end if

  if (use_mask .eq. 1) then
    if (ltrace) print *,' allocating imask'
    allocate(imask(nx, ny, nz),STAT=istat)
    if (istat .ne. 0) then
      write(0,*)'Could not allocate imask in HyperAnalysis'
      write(0,*)'   nx = ',nx
      write(0,*)'   ny = ',ny
      write(0,*)'   nz = ',nz
      stop
    end if

    if (ltrace) print *,' stat = ',istat

    if (order .eq. 42) then
      call SetImask42(w, imask, par)
    else
      call set_mask(w, imask, par)
    end if    
    
  end if

  call hyperauxvars(v, w, u2, par)  
  
  if (nint(par(P_POINT_WISE_ANALYSIS)) .eq. 0) then 
  
    if (P_POINT_WISE_DERIVATIVES .eq. 0) then
      if (ltrace) then
        write(0,*)'>>> HYPER_ANALYSIS: calling hyperderivs'
      end if
  !   This seems unnecessary for now, and very expensive.  I will
  !   comment this out.  ---Dave
      call hyperderivs(dxu,   dyu,  dzu,  dxv,  dyv,  dzv, &
                       u2,    v,    w,     imask,      par)
    end if

    if (ltrace) then
      write(0,*)'>>> HYPER_ANALYSIS: calling analysis'
    end if
    
    call analysis(u2, u0, dxu, dyu, dzu, v, dxv, dyv, dzv, w, par)
  else
  
    if (ltrace) then
      write(0,*)'>>> HYPER_ANALYSIS: calling hyperderivs and analysis point to point'
    end if
 
    !tempimask = 1 

    ! Avoid using ghostzones
    if ( par(P_BBOX1) .eq. 1 ) then
      ilo = 1
    else
      ilo = ng+1
    end if

    if ( par(P_BBOX2) .eq. 1 ) then
      ihi = nx
    else
      ihi = nx-ng-1
    end if

    if ( par(P_BBOX3) .eq. 1 ) then
      jlo = 1
    else
      jlo = ng+1
    end if

    if ( par(P_BBOX4) .eq. 1 ) then
      jhi = ny
    else
      jhi = ny-ng-1
    end if

    if ( par(P_BBOX5) .eq. 1 ) then
      klo = 1
    else
      klo = ng+1
    end if

    if ( par(P_BBOX6) .eq. 1 ) then
      khi = nz
    else
      khi = nz-ng-1
    end if

!   do kk = klo, khi
!     do jj = jlo, jhi
!       do ii = ilo, ihi
       
     do kk = 1, nz
       do jj = 1, ny
         do ii = 1, nx

          if ((w(H_MASK)%d(ii,jj,kk).gt.0) .OR. (use_mask .eq. 0) ) then
  
            do m = 1, NU
              u_pt(m) = u2(m)%d(ii,jj,kk)
            end do
            do m = 1, NV
              v_pt(m) = v(m)%d(ii,jj,kk)
            end do
            do m = 1, NW
              w_pt(m) = w(m)%d(ii,jj,kk)
            end do
            if (use_mask .eq. 1) then
               ! This applies if imask() was allocated
               imask_pt = imask(ii,jj,kk)
            else
               ! Just a dummy value:
               imask_pt = 1
            end if

            call hyperderivs_pt(dxu_pt,   dyu_pt,  dzu_pt,                &
                                dxv_pt,   dyv_pt,  dzv_pt,                &
                                u2,       v,       w_pt,    imask_pt,     & 
                                ii,       jj,      kk,      par           )

            call analysis_pt(u_pt, u_pt, dxu_pt, dyu_pt, dzu_pt, &
                             v_pt, dxv_pt, dyv_pt, dzv_pt, w_pt, par)      
			       
            do m=1, NV
              v(m)%d(ii,jj,kk) = v_pt(m)
            end do
            do m = 1, NW
              w(m)%d(ii,jj,kk) = w_pt(m)
            end do
 
	  end if
	  
	    
        end do
      end do
    end do
    
  end if      

  if (ltrace) then
     write(0,*)'>>> HYPER_ANALYSIS: return from analysis'
  end if
  
  if (use_mask .eq. 1) then
    if (ltrace) print *,' about to deallocate imask'
    if (allocated(imask)) then
      deallocate(imask)
    else
      write(0,*)'>>> HYPER_ANALYSIS: imask not allocated as expected!'
    end if
    if (ltrace) print *,' end of deallocate imask'
  end if
  
  if (ltrace) then
     write(0,*)'>>> HYPER_ANALYSIS: exit from HyperAnalysis'
  end if

  return
  end subroutine HyperAnalysis
    
  
  
