
#include "cctk.h"

MODULE HYPER_DIS_21
use params

CONTAINS

!-----------------------------------------------------------------------
!
!    SUBROUTINE Dis21NoMask:
!
!    implements Dissipation.  2nd order interior, 1st order boundary
!    reuses the space allocated for the standard derivatives
!
!-----------------------------------------------------------------------
  subroutine Dis21NoMask(Du, u, direction, par)

    CCTK_REAL, DIMENSION(:), INTENT(IN)      :: par
    CCTK_REAL, DIMENSION(:,:,:), INTENT(IN)  :: u
    CCTK_REAL, DIMENSION(:,:,:), INTENT(OUT) :: Du
    CCTK_INT :: direction


    ! local vars
    CCTK_REAL :: dx, dy, dz
    CCTK_INT  :: i, j, k,nx, ny, nz, shp(3)
    

    shp = shape(u)
    nx = shp(1)
    ny = shp(2)
    nz = shp(3)

    dx = par(P_DX)
    dy = par(P_DY)
    dz = par(P_DZ)
    
    if (direction.eq.1) then 
      Du(1,:,:) = 2.*(u(2,:,:)-u(1,:,:))
      do i = 2, nx-1
        Du(i,:,:) = u(i+1,:,:)-2.*u(i,:,:)+u(i-1,:,:)
      end do
      Du(nx,:,:) = -2.*(u(nx,:,:)-u(nx-1,:,:))
    else if (direction.eq.2) then 
      Du(:,1,:) = 2.*(u(:,2,:)-u(:,1,:))
      do j = 2, ny-1
        Du(:,j,:) = u(:,j+1,:)-2.*u(:,j,:)+u(:,j-1,:)
      end do
      Du(:,ny,:) = -2.*(u(:,ny,:)-u(:,ny-1,:))
    else if (direction==3) then 
      Du(:,:,1) = 2.*(u(:,:,2)-u(:,:,1))
      do k = 2, nz-1
        Du(:,:,k) = u(:,:,k+1)-2.*u(:,:,k)+u(:,:,k-1)
      end do
      Du(:,:,nz) = -2.*(u(:,:,nz)-u(:,:,nz-1))
    else 
      write(*,*) "D21NoMask: Wrong direction in the derivative" 
      STOP
    end if 

    return
  end subroutine Dis21NoMask


!-----------------------------------------------------------------------
!
!    Dis21
!
!-----------------------------------------------------------------------
  subroutine Dis21(Du, u, direction, imask, par)
    implicit none

    CCTK_REAL, DIMENSION(:), INTENT(IN)      :: par
    CCTK_REAL, DIMENSION(:,:,:), INTENT(IN)  :: u
    CCTK_REAL, DIMENSION(:,:,:), INTENT(OUT) :: Du
    CCTK_INT                                 :: direction
    CCTK_INT, dimension(:,:,:)               :: imask
    CCTK_INT                                 :: d


    ! local vars
    CCTK_REAL :: dx, dy, dz
    CCTK_INT  :: i, j, k,nx, ny, nz, shp(3)

    !  If ltrace2 is TRUE, then a warning message is printed for 
    !  undefined derivatives are flagged.
    !  If ltrace2 is FALSE, then no warning message is printed.
    logical, parameter :: ltrace2 = .false.


    shp = shape(u)
    nx = shp(1)
    ny = shp(2)
    nz = shp(3)

    dx = par(P_DX)
    dy = par(P_DY)
    dz = par(P_DZ)

    if (direction.eq.1) then
      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = imask(i,j,k) / 10000
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)
        else if (d .eq. P_STENCIL_CENTER_RIGHT .or. &
                 d .eq. P_STENCIL_CENTER_LEFT) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = 2.0*(u(i+1,j,k) - u(i,j,k))
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0*(u(i,j,k) - u(i-1,j,k))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)   ! not correct 
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = u(i+1,j,k)-2.0*u(i,j,k)+u(i-1,j,k)   ! not correct
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'diss21: Illegal x derivative type ',d
            write(0,*)'diss21: imask                     ',imask(i,j,k)
            write(0,*)'diss21: at (i,j,k) location: ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do

    else if (direction.eq.2) then

      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = mod((imask(i,j,k)/100),100)
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = 2.0*(u(i,j+1,k) - u(i,j,k))
        else if (d .eq. P_STENCIL_CENTER_LEFT .or. &
                 d .eq. P_STENCIL_CENTER_RIGHT) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0*(u(i,j,k) - u(i,j-1,k))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)   ! not correct
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = u(i,j+1,k) -2.0*u(i,j,k) + u(i,j-1,k)   ! not correct
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'diss21: Illegal y derivative type ',d
            write(0,*)'diss21: imask                     ',imask(i,j,k)
            write(0,*)'diss21: at (i,j,k) location ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do

    else if (direction.eq.3) then
 
      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = mod(imask(i,j,k),100)
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = 2.0*(u(i,j,k+1) - u(i,j,k))
        else if (d .eq. P_STENCIL_CENTER_LEFT .or. &
                 d .eq. P_STENCIL_CENTER_RIGHT) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0*(u(i,j,k) - u(i,j,k-1))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)   ! not correct
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)   ! not correct
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = u(i,j,k+1) - 2.0*u(i,j,k) + u(i,j,k-1)   ! not correct
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'diss21: Illegal z derivative type ',d
            write(0,*)'diss21: imask                     ',imask(i,j,k)
            write(0,*)'diss21: at (i,j,k) location ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do

    else
      write(0,*)'d21:  unknown direction ',direction
    end if

    return
  end subroutine Dis21

!-----------------------------------------------------------------------
!
!    SUBROUTINE KODis21NoMask:
!
!    implements Kreiss-Oliger Dissipation.
!    reuses the space allocated for the standard derivatives
!
!-----------------------------------------------------------------------
  subroutine KODis21NoMask(Du, u, direction, par)

    CCTK_REAL, DIMENSION(:), INTENT(IN)      :: par
    CCTK_REAL, DIMENSION(:,:,:), INTENT(IN)  :: u
    CCTK_REAL, DIMENSION(:,:,:), INTENT(OUT) :: Du
    CCTK_INT :: direction


    ! local vars
    CCTK_REAL :: dx, dy, dz
    CCTK_INT  :: i, j, k,nx, ny, nz, shp(3)


    shp = shape(u)
    nx = shp(1)
    ny = shp(2)
    nz = shp(3)

    dx = par(P_DX)
    dy = par(P_DY)
    dz = par(P_DZ)

    if (direction.eq.1) then
       Du(1,:,:) = -2.0 * (u(3,:,:) - 2.0*u(2,:,:) + u(1,:,:)) / dx

       Du(2,:,:) = - (u(4,:,:) - 4.0*u(3,:,:) + 5.0*u(2,:,:) &
                                - 2.0*u(1,:,:)) /dx

       do i = 3, nx-2
          Du(i,:,:) = - (u(i+2,:,:) - 4.0*u(i+1,:,:) + 6.0*u(i,:,:) &
                             - 4.0*u(i-1,:,:) + u(i-2,:,:)) / dx
       end do

       Du(nx-1,:,:) = - (u(nx-3,:,:) - 4.0*u(nx-2,:,:) + 5.0*u(nx-1,:,:) &
                                - 2.0*u(nx,:,:)) / dx

       Du(nx,:,:) = -2.0 * (u(nx-2,:,:) - 2.0*u(nx-1,:,:) + u(nx,:,:)) / dx

    else if (direction.eq.2) then
       Du(:,1,:) = -2.0 * (u(:,3,:) - 2.0*u(:,2,:) + u(:,1,:)) / dy

       Du(:,2,:) = - (u(:,4,:) - 4.0*u(:,3,:) + 5.0*u(:,2,:) &
                                - 2.0*u(:,1,:)) /dy

       do j = 3, ny-2
          Du(:,j,:) = - (u(:,j+2,:) - 4.0*u(:,j+1,:) + 6.0*u(:,j,:) &
                             - 4.0*u(:,j-1,:) + u(:,j-2,:)) / dy
       end do

       Du(:,ny-1,:) = - (u(:,ny-3,:) - 4.0*u(:,ny-2,:) + 5.0*u(:,ny-1,:) &
                                - 2.0*u(:,ny,:)) / dy

       Du(:,ny,:) = -2.0 * (u(:,ny-2,:) - 2.0*u(:,ny-1,:) + u(:,ny,:)) / dy


    else if (direction==3) then

       Du(:,:,1) = -2.0 * (u(:,:,3) - 2.0*u(:,:,2) + u(:,:,1)) / dz

       Du(:,:,2) = - (u(:,:,4) - 4.0*u(:,:,3) + 5.0*u(:,:,2) &
                                - 2.0*u(:,:,1)) /dz

       do k = 3, nz-2
          Du(:,:,k) = - (u(:,:,k+2) - 4.0*u(:,:,k+1) + 6.0*u(:,:,k) &
                             - 4.0*u(:,:,k-1) + u(:,:,k-2)) / dz
       end do

       Du(:,:,nz-1) = - (u(:,:,nz-3) - 4.0*u(:,:,nz-2) + 5.0*u(:,:,nz-1) &
                                - 2.0*u(:,:,nz)) / dz

       Du(:,:,nz) = -2.0 * (u(:,:,nz-2) - 2.0*u(:,:,nz-1) + u(:,:,nz)) / dz

    else
      write(*,*) "D21NoMask: Wrong direction in the derivative"
      STOP
    end if

    return
  end subroutine KODis21NoMask


!-----------------------------------------------------------------------
!
!    KODis21
!
!-----------------------------------------------------------------------
  subroutine KODis21(Du, u, direction, imask, par)
    implicit none

    CCTK_REAL, DIMENSION(:), INTENT(IN)      :: par
    CCTK_REAL, DIMENSION(:,:,:), INTENT(IN)  :: u
    CCTK_REAL, DIMENSION(:,:,:), INTENT(OUT) :: Du
    CCTK_INT                                 :: direction
    CCTK_INT, dimension(:,:,:)               :: imask
    CCTK_INT                                 :: d


    ! local vars
    CCTK_REAL :: dx, dy, dz, c
    CCTK_INT  :: i, j, k,nx, ny, nz, shp(3)

    !  If ltrace2 is TRUE, then a warning message is printed for
    !  undefined derivatives are flagged.
    !  If ltrace2 is FALSE, then no warning message is printed.
    logical, parameter :: ltrace2 = .true.


    shp = shape(u)
    nx = shp(1)
    ny = shp(2)
    nz = shp(3)

    dx = par(P_DX)
    dy = par(P_DY)
    dz = par(P_DZ)


    if (direction.eq.1) then
      c = 1.0/dx
      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = imask(i,j,k) / 10000
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = -c * (u(i+2,j,k) - 4.0*u(i+1,j,k) + 6.0*u(i,j,k) &
                                - 4.0*u(i-1,j,k) + u(i-2,j,k))
        else if (d .eq. P_STENCIL_CENTER_RIGHT) then
          Du(i,j,k) = -c * (u(i-2,j,k) - 4.0*u(i-1,j,k) + 5.0*u(i,j,k) &
                                - 2.0*u(i+1,j,k))
        else if (d .eq. P_STENCIL_CENTER_LEFT) then
          Du(i,j,k) = -c * (u(i+2,j,k) - 4.0*u(i+1,j,k) + 5.0*u(i,j,k) &
                                - 2.0*u(i-1,j,k))
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = -2.0 * c * (u(i+2,j,k) - 2.0*u(i+1,j,k) + u(i,j,k))
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0 * c * (u(i-2,j,k) - 2.0*u(i-1,j,k) + u(i,j,k))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = -2.0/3.0*c* (2.0*u(i+2,j,k) - 4.0*u(i+1,j,k) &
                             + 3.0*u(i,j,k) - 2.0*u(i-1,j,k) + u(i-2,j,k))
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = -2.0/3.0*c* (u(i+2,j,k) - 2.0*u(i+1,j,k) &
                             + 3.0*u(i,j,k) - 4.0*u(i-1,j,k) + 2.0*u(i-2,j,k))
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = -2.0/7.0*c * (4.0*u(i+2,j,k) - 8.0*u(i+1,j,k)  &
                             + 7.0*u(i,j,k) - 6.0*u(i-1,j,k) + 3.0*u(i-2,j,k))
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = -2.0/7.0*c * (3.0*u(i+2,j,k) - 6.0*u(i+1,j,k)  &
                             + 7.0*u(i,j,k) - 8.0*u(i-1,j,k) + 4.0*u(i-2,j,k))
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'KOdis21: invalid x derivative type ',d
            write(0,*)'     at location ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do
    end if

    if (direction.eq.2) then
      c = 1.0/dy
      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = mod((imask(i,j,k)/100),100)
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = -c * (u(i,j+2,k) - 4.0*u(i,j+1,k) + 6.0*u(i,j,k) &
                                - 4.0*u(i,j-1,k) + u(i,j-2,k))
        else if (d .eq. P_STENCIL_CENTER_RIGHT) then
          Du(i,j,k) = -c * (u(i,j-2,k) - 4.0*u(i,j-1,k) + 5.0*u(i,j,k) &
                                - 2.0*u(i,j+1,k))
        else if (d .eq. P_STENCIL_CENTER_LEFT) then
          Du(i,j,k) = -c * (u(i,j+2,k) - 4.0*u(i,j+1,k) + 5.0*u(i,j,k) &
                                - 2.0*u(i,j-1,k))
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = -2.0 * c * (u(i,j+2,k) - 2.0*u(i,j+1,k) + u(i,j,k))
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0 * c * (u(i,j-2,k) - 2.0*u(i,j-1,k) + u(i,j,k))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = -2.0/3.0*c* (2.0*u(i,j+2,k) - 4.0*u(i,j+1,k) &
                             + 3.0*u(i,j,k) - 2.0*u(i,j-1,k) + u(i,j-2,k))
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = -2.0/3.0*c* (u(i,j+2,k) - 2.0*u(i,j+1,k) &
                             + 3.0*u(i,j,k) - 4.0*u(i,j-1,k) + 2.0*u(i,j-2,k))
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = -2.0/7.0*c * (4.0*u(i,j+2,k) - 8.0*u(i,j+1,k)  &
                             + 7.0*u(i,j,k) - 6.0*u(i,j-1,k) + 3.0*u(i,j-2,k))
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = -2.0/7.0*c * (3.0*u(i,j+2,k) - 6.0*u(i,j+1,k)  &
                             + 7.0*u(i,j,k) - 8.0*u(i,j-1,k) + 4.0*u(i,j-2,k))
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'KOdis21: invalid y derivative type ',d
            write(0,*)'     at location ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do
    end if

    if (direction.eq.3) then
      c = 1.0/dz
      do k = 1, nz
      do j = 1, ny
      do i = 1, nx
        d = mod(imask(i,j,k),100)
        if (d .eq. P_STENCIL_CENTER) then
          Du(i,j,k) = -c * (u(i,j,k+2) - 4.0*u(i,j,k+1) + 6.0*u(i,j,k) &
                                - 4.0*u(i,j,k-1) + u(i,j,k-2))
        else if (d .eq. P_STENCIL_CENTER_RIGHT) then
          Du(i,j,k) = -c * (u(i,j,k-2) - 4.0*u(i,j,k-1) + 5.0*u(i,j,k) &
                                - 2.0*u(i,j,k+1))
        else if (d .eq. P_STENCIL_CENTER_LEFT) then
          Du(i,j,k) = -c * (u(i,j,k+2) - 4.0*u(i,j,k+1) + 5.0*u(i,j,k) &
                                - 2.0*u(i,j,k-1))
        else if (d .eq. P_STENCIL_LEFT) then
          Du(i,j,k) = -2.0 * c * (u(i,j,k+2) - 2.0*u(i,j,k+1) + u(i,j,k))
        else if (d .eq. P_STENCIL_RIGHT) then
          Du(i,j,k) = -2.0 * c * (u(i,j,k-2) - 2.0*u(i,j,k-1) + u(i,j,k))
        else if (d .eq. P_STENCIL_MASK) then
          Du(i,j,k) = 0.0
        else if (d .eq. P_STENCIL_EDGE_P) then
          Du(i,j,k) = -2.0/3.0*c* (2.0*u(i,j,k+2) - 4.0*u(i,j,k+1) &
                             + 3.0*u(i,j,k) - 2.0*u(i,j,k-1) + u(i,j,k-2))
        else if (d .eq. P_STENCIL_EDGE_M) then
          Du(i,j,k) = -2.0/3.0*c* (u(i,j,k+2) - 2.0*u(i,j,k+1) &
                             + 3.0*u(i,j,k) - 4.0*u(i,j,k-1) + 2.0*u(i,j,k-2))
        else if (d .eq. P_STENCIL_VERTEX_P) then
          Du(i,j,k) = -2.0/7.0*c * (4.0*u(i,j,k+2) - 8.0*u(i,j,k+1)  &
                             + 7.0*u(i,j,k) - 6.0*u(i,j,k-1) + 3.0*u(i,j,k-2))
        else if (d .eq. P_STENCIL_VERTEX_M) then
          Du(i,j,k) = -2.0/7.0*c * (3.0*u(i,j,k+2) - 6.0*u(i,j,k+1)  &
                             + 7.0*u(i,j,k) - 8.0*u(i,j,k-1) + 4.0*u(i,j,k-2))
        else if (d .eq. P_STENCIL_UNDEF) then
          Du(i,j,k) = 0.0
        else
          if (ltrace2) then
            write(0,*)'KOdis21: invalid z derivative type ',d
            write(0,*)'     at location ',i,j,k
          end if
          Du(i,j,k) = 0.0
        end if
      end do
      end do
      end do
    end if




    return
  end subroutine KODis21


END MODULE HYPER_DIS_21
