!-----------------------------------------------------------------
!
!  $Id: hyperderivs.f90,v 1.11 2006/09/13 15:51:54 steve Exp $
!
!-----------------------------------------------------------------

#include "cctk.h"

MODULE HYPER_DERIVS

  use params
  use GF
  use HYPER_DERIVS_21
  use HYPER_DERIVS_21_PT
  use HYPER_DERIVS_42  
  use HYPER_DERIVS_42_PT  
  use HYPER_DERIVS_43
  use HYPER_DERIVS_43_PT
 
contains

!-----------------------------------------------------------------
!
! SUBROUTINE HYPERDERIVS
!
!-----------------------------------------------------------------
  subroutine hyperderivs(dxu, dyu, dzu, dxv, dyv, dzv, &
                         u, v, w, imask, par)
    implicit none
 
    type(gridfunction), dimension(:)   :: dxu,dyu,dzu,u
    type(gridfunction), dimension(:)   :: dxv,dyv,dzv,v
    type(gridfunction), dimension(NW)   :: w
    CCTK_INT, dimension(:,:,:)          :: imask
    CCTK_REAL, dimension(:)             :: par

    ! local vars
    CCTK_INT                            :: direc, m, use_mask, order
    CCTK_INT                            :: lnu, lnv
    CCTK_REAL                           :: normx, normy, normz
    CCTK_REAL, pointer, dimension(:,:,:):: mask, x, y, z
    target                              :: w
    CCTK_INT                            :: nx, ny, nz, myid
    logical, parameter                  :: ltrace = .false.
    CCTK_REAL                           :: myl2norm3d
    CCTK_INT                            :: proc_return_myid



    if (ltrace) then
      myid = proc_return_myid()
      nx   = nint(par(P_NX))
      ny   = nint(par(P_NY))
      nz   = nint(par(P_NZ))
      write(0,*)myid,'] ************** hyperderivs *******************'
    end if

    lnu = size(u)
    lnv = size(v)
     
    use_mask = nint(par(P_USE_MASK))
    order    = nint(par(P_DERIV_ORDER))

    mask => w(H_MASK)%d
    x    => w(H_X)%d
    y    => w(H_Y)%d
    z    => w(H_Z)%d

    if (use_mask .eq. 1) then
      if (ltrace) write(0,*)myid, '] >>> HYPERDERIVS: use_mask == 1'
      !---------------------------------------
      ! MASKED DERIVATIVES
      !---------------------------------------
      if (order .eq. 2) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          direc = 1
          if (u(m)%take_dx .eq. 1) then
            call D21(dxu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 2
          if (u(m)%take_dy .eq. 1) then
            call D21(dyu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 3
          if (u(m)%take_dz .eq. 1) then
            call D21(dzu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          direc = 1
          if (v(m)%take_dx .eq. 1) then
            call D21(dxv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 2
          if (v(m)%take_dy .eq. 1) then
            call D21(dyv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 3
          if (v(m)%take_dz .eq. 1) then
            call D21(dzv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
          end if
        end do

      else if (order .eq. 42) then   
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          direc = 1
          if (u(m)%take_dx .eq. 1) then
            call D42(dxu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 2
          if (u(m)%take_dy .eq. 1) then
            call D42(dyu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 3
          if (u(m)%take_dz .eq. 1) then
            call D42(dzu(m)%d,u(m)%d,direc,x,y,z,mask,imask,par)
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          direc = 1
          if (v(m)%take_dx .eq. 1) then
            call D42(dxv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
          end if

          direc = 2
         if (v(m)%take_dy .eq. 1) then
            call D42(dyv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
         end if

          direc = 3
         if (v(m)%take_dz .eq. 1) then
            call D42(dzv(m)%d,v(m)%d,direc,x,y,z,mask,imask,par)
          end if
        end do	
		
      else
        write(0,*)'A: unmasked derivatives not implemented for order ',order
        stop
      end if
    else
      if (ltrace) write(0,*)myid,'] >>> HYPERDERIVS: use_mask != 1'
      !---------------------------------------
      ! UNMASKED DERIVATIVES
      !---------------------------------------
      if (order .eq. 2) then
        if (ltrace) write(0,*)myid,'] >>> HYPERDERIVS: order == 2, derivs of U:',lnu
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          direc = 1
          if (u(m)%take_dx .eq. 1) then
            call D21NoMask(dxu(m)%d,u(m)%d,direc,par)
          end if

          direc = 2
          if (u(m)%take_dy .eq. 1) then
            call D21NoMask(dyu(m)%d,u(m)%d,direc,par)
          end if

          direc = 3
          if (u(m)%take_dz .eq. 1) then
            call D21NoMask(dzu(m)%d,u(m)%d,direc,par)
          end if
        end do
        !
        ! Derivatives of V
        !
        if (ltrace) write(0,*)myid,'] >>> HYPERDERIVS:             derivs of V:',lnv
        do m = 1, lnv
          direc = 1
          if (v(m)%take_dx .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: x-deriv for v',m
            call D21NoMask(dxv(m)%d,v(m)%d,direc,par)
          end if

          direc = 2
          if (v(m)%take_dy .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: y-deriv for v',m
            call D21NoMask(dyv(m)%d,v(m)%d,direc,par)
          end if

          direc = 3
          if (v(m)%take_dz .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: z-deriv for v',m
            call D21NoMask(dzv(m)%d,v(m)%d,direc,par)
          end if
        end do

!4th - 3rd order
     else if (order .eq. 4) then
        if (ltrace) write(0,*)'>>> HYPERDERIVS: order == 2, derivs of U:',lnu
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          direc = 1
          if (u(m)%take_dx .eq. 1) then
            call D43NoMask(dxu(m)%d,u(m)%d,direc,par)
          end if

          direc = 2
          if (u(m)%take_dy .eq. 1) then
            call D43NoMask(dyu(m)%d,u(m)%d,direc,par)
          end if

          direc = 3
          if (u(m)%take_dz .eq. 1) then
            call D43NoMask(dzu(m)%d,u(m)%d,direc,par)
          end if
        end do
        !
        ! Derivatives of V
        !
        if (ltrace) write(0,*)'>>> HYPERDERIVS:             derivs of V:',lnv
        do m = 1, lnv
          direc = 1
          if (v(m)%take_dx .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: x-deriv for v',m
            call D43NoMask(dxv(m)%d,v(m)%d,direc,par)
          end if

          direc = 2
          if (v(m)%take_dy .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: y-deriv for v',m
            call D43NoMask(dyv(m)%d,v(m)%d,direc,par)
          end if

          direc = 3
          if (v(m)%take_dz .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: z-deriv for v',m
            call D43NoMask(dzv(m)%d,v(m)%d,direc,par)
          end if
        end do

	
!4th - 2rd order
     else if (order .eq. 42) then
        if (ltrace) write(0,*)'>>> HYPERDERIVS: order == 2, derivs of U:',lnu
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          direc = 1
          if (u(m)%take_dx .eq. 1) then
            call D42NoMask(dxu(m)%d,u(m)%d,direc,par)
          end if

          direc = 2
          if (u(m)%take_dy .eq. 1) then
            call D42NoMask(dyu(m)%d,u(m)%d,direc,par)
          end if

          direc = 3
          if (u(m)%take_dz .eq. 1) then
            call D42NoMask(dzu(m)%d,u(m)%d,direc,par)
          end if
        end do
        !
        ! Derivatives of V
        !
        if (ltrace) write(0,*)'>>> HYPERDERIVS:             derivs of V:',lnv
        do m = 1, lnv
          direc = 1
          if (v(m)%take_dx .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: x-deriv for v',m
            call D42NoMask(dxv(m)%d,v(m)%d,direc,par)
          end if

          direc = 2
          if (v(m)%take_dy .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: y-deriv for v',m
            call D42NoMask(dyv(m)%d,v(m)%d,direc,par)
          end if

          direc = 3
          if (v(m)%take_dz .eq. 1) then
            if (ltrace) write(0,*)'>>> HYPERDERIVS: z-deriv for v',m
            call D42NoMask(dzv(m)%d,v(m)%d,direc,par)
          end if
        end do
	
      else 
        write(0,*)'B: unmasked derivatives not implemented for order ',order
        stop
      end if
    end if

  if (ltrace) then
     write(0,*) myid,'] >>>Hyperderivs: Norms aftewards:'
     do m=1,lNU
        if (u(m)%take_dx .eq. 1) then
           write(0,*) myid,'] Der: ||u_x||  ',m,myl2norm3d(dxu(m)%d, nx, ny, nz)
        end if
        if (u(m)%take_dy .eq. 1) then
           write(0,*) myid,'] Der: ||u_y||  ',m,myl2norm3d(dyu(m)%d, nx, ny, nz)
        end if
        if (u(m)%take_dz .eq. 1) then
           write(0,*) myid,'] Der: ||u_z||  ',m,myl2norm3d(dzu(m)%d, nx, ny, nz)
        end if
     end do
     do m=1,lNV
        if (v(m)%take_dx .eq. 1) then
           write(0,*) myid,'] Der: ||v_x||  ',m,myl2norm3d(dxv(m)%d, nx, ny, nz)
        end if
        if (v(m)%take_dy .eq. 1) then
           write(0,*) myid,'] Der: ||v_y||  ',m,myl2norm3d(dyv(m)%d, nx, ny, nz)
        end if
        if (v(m)%take_dz .eq. 1) then
           write(0,*) myid,'] Der: ||v_z||  ',m,myl2norm3d(dzv(m)%d, nx, ny, nz)
        end if
     end do
     write(0,*)myid,'] >>> HYPERDERIVS: Done.'
  end if

    return
  end subroutine hyperderivs


!-----------------------------------------------------------------
!
! SUBROUTINE HYPERDERIVS_PT
!
! Point-wise derivatives.
!
!-----------------------------------------------------------------
  subroutine hyperderivs_pt(dxu, dyu, dzu, dxv, dyv, dzv, &
                            u, v, wpt, imask, i, j, k, par)
    implicit none
 
    CCTK_REAL, dimension(:)   :: dxu, dyu, dzu
    CCTK_REAL, dimension(:)   :: dxv, dyv, dzv
    type(gridfunction), dimension(:)   :: u
    type(gridfunction), dimension(:)   :: v
    CCTK_REAL, dimension(NW)            :: wpt
    CCTK_INT                            :: imask, point_wise_analysis
    CCTK_REAL, dimension(:)             :: par
    CCTK_INT                            :: i, j, k

    CCTK_INT                            :: nx, ny, nz
    CCTK_INT                            :: lnu, lnv
    CCTK_INT                            :: m, use_mask, order
    CCTK_REAL, parameter                :: ERROR_VALUE = 1.0d200
    logical, parameter                  :: ltrace = .false.

    if (ltrace) then
      write(0,*)'Entering hyperderivs_pt'
    end if

    lnu = size(u)
    lnv = size(v)

    use_mask = nint(par(P_USE_MASK))
    order    = nint(par(P_DERIV_ORDER))
    point_wise_analysis = nint(par(P_POINT_WISE_ANALYSIS))

    nx = nint(par(P_NX))
    ny = nint(par(P_NY))
    nz = nint(par(P_NZ))

    if (i .lt. 1 .or. i .gt. nx) then
      write(0,*)'HYPERDERIVS_PT >> i out of range: i/nx = ',i,nx
    end if
    if (j .lt. 1 .or. j .gt. ny) then
      write(0,*)'HYPERDERIVS_PT >> j out of range: y/ny = ',j,ny
    end if
    if (k .lt. 1 .or. k .gt. nz) then
      write(0,*)'HYPERDERIVS_PT >> k out of range: k/nz = ',k,nz
    end if

    if (use_mask .eq. 1) then
      !---------------------------------------
      ! MASKED DERIVATIVES FOR 2-1 
      !---------------------------------------
      if (order .eq. 2) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          if (u(m)%take_dx .eq. 1 .or. u(m)%take_dy .eq. 1  &
                                  .or. u(m)%take_dz .eq. 1  &
	                          .or. point_wise_analysis .eq. 1) then
            call D21_pt(dxu(m), dyu(m), dzu(m), u(m)%d,   &
                        wpt,    imask,  i,      j,      k,      par)
          else
            dxu(m) = ERROR_VALUE
            dyu(m) = ERROR_VALUE
            dzu(m) = ERROR_VALUE
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          if (v(m)%take_dx .eq. 1 .or. v(m)%take_dy .eq. 1  &
                                  .or. v(m)%take_dz .eq. 1) then
            call D21_pt(dxv(m),  dyv(m),  dzv(m),  v(m)%d,  &
                        wpt,     imask,   i,       j,       k,    par)
          else
            dxv(m) = ERROR_VALUE
            dyv(m) = ERROR_VALUE
            dzv(m) = ERROR_VALUE
          end if
        end do


      !---------------------------------------
      ! MASKED DERIVATIVES FOR 4-2
      !---------------------------------------
      else if (order .eq. 42) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          if (u(m)%take_dx .eq. 1 .or. u(m)%take_dy .eq. 1  &
                                  .or. u(m)%take_dz .eq. 1  &
	                          .or. point_wise_analysis .eq. 1) then
            call D42_pt(dxu(m), dyu(m), dzu(m), u(m)%d,   &
                        wpt,    imask,  i,      j,      k,      par)
          else
            dxu(m) = ERROR_VALUE
            dyu(m) = ERROR_VALUE
            dzu(m) = ERROR_VALUE
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          if (v(m)%take_dx .eq. 1 .or. v(m)%take_dy .eq. 1  &
                                  .or. v(m)%take_dz .eq. 1) then
            call D42_pt(dxv(m),  dyv(m),  dzv(m),  v(m)%d,  &
                        wpt,     imask,   i,       j,       k,    par)
          else
            dxv(m) = ERROR_VALUE
            dyv(m) = ERROR_VALUE
            dzv(m) = ERROR_VALUE
          end if
        end do
	
	
      else
        write(0,*)'C: masked point-wise derivatives not implemented'
        write(0,*)'   for order ',order
        stop
      end if
    else
      !---------------------------------------
      ! UNMASKED DERIVATIVES FOR 2-1
      !---------------------------------------
      if (order .eq. 2) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          if (u(m)%take_dx .eq. 1 .or. u(m)%take_dy .eq. 1  &
                                  .or. u(m)%take_dz .eq. 1  &
                                  .or. point_wise_analysis .eq. 1) then
            call D21NoMask_pt(dxu(m),dyu(m),dzu(m),u(m)%d,wpt,i,j,k,par)
          else
            dxu(m) = ERROR_VALUE
            dyu(m) = ERROR_VALUE
            dzu(m) = ERROR_VALUE
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          if (v(m)%take_dx .eq. 1 .or. v(m)%take_dy .eq. 1   &
                                  .or. v(m)%take_dz .eq. 1) then
            call D21NoMask_pt(dxv(m),dyv(m),dzv(m),v(m)%d,wpt,i,j,k,par)
          else
            dxv(m) = ERROR_VALUE
            dyv(m) = ERROR_VALUE
            dzv(m) = ERROR_VALUE
          end if
        end do

      !---------------------------------------
      ! UNMASKED DERIVATIVES FOR 4-3
      !---------------------------------------	
      else  if (order .eq. 4) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          if (u(m)%take_dx .eq. 1 .or. u(m)%take_dy .eq. 1  &
                                  .or. u(m)%take_dz .eq. 1  &
                                  .or. point_wise_analysis .eq. 1) then
            call D43NoMask_pt(dxu(m),dyu(m),dzu(m),u(m)%d,i,j,k,par)
          else
            dxu(m) = ERROR_VALUE
            dyu(m) = ERROR_VALUE
            dzu(m) = ERROR_VALUE
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          if (v(m)%take_dx .eq. 1 .or. v(m)%take_dy .eq. 1   &
                                  .or. v(m)%take_dz .eq. 1) then
            call D43NoMask_pt(dxv(m),dyv(m),dzv(m),v(m)%d,i,j,k,par)
          else
            dxv(m) = ERROR_VALUE
            dyv(m) = ERROR_VALUE
            dzv(m) = ERROR_VALUE
          end if
        end do

      !---------------------------------------
      ! UNMASKED DERIVATIVES FOR 4-2
      !---------------------------------------	
      else  if (order .eq. 42) then
        !
        ! Derivatives of U
        !
        do m = 1, lnu
          if (u(m)%take_dx .eq. 1 .or. u(m)%take_dy .eq. 1  &
                                  .or. u(m)%take_dz .eq. 1  &
                                  .or. point_wise_analysis .eq. 1) then
            call D42NoMask_pt(dxu(m),dyu(m),dzu(m),u(m)%d,wpt,i,j,k,par)
          else
            dxu(m) = ERROR_VALUE
            dyu(m) = ERROR_VALUE
            dzu(m) = ERROR_VALUE
          end if
        end do
        !
        ! Derivatives of V
        !
        do m = 1, lnv
          if (v(m)%take_dx .eq. 1 .or. v(m)%take_dy .eq. 1   &
                                  .or. v(m)%take_dz .eq. 1) then
            call D42NoMask_pt(dxv(m),dyv(m),dzv(m),v(m)%d,wpt,i,j,k,par)
          else
            dxv(m) = ERROR_VALUE
            dyv(m) = ERROR_VALUE
            dzv(m) = ERROR_VALUE
          end if
        end do
      
      
      else 
        write(0,*)'D: unmasked point-wise derivatives not implemented '
        write(0,*)'   for order ',order
        stop
      end if
    end if

    return
  end subroutine hyperderivs_pt

end MODULE HYPER_DERIVS

