!-----------------------------------------------------------------
!
!  $Id: hyperrhs.f90,v 1.14 2006/09/13 15:51:54 steve Exp $
!
!-----------------------------------------------------------------

#include "cctk.h"

MODULE HYPER_RHS
    use params
    use GF
    use rhs_mine
    use HYPER_DERIVS
    use HYPER_AUXVARS
  contains

!-----------------------------------------------------------------
!
! SUBROUTINE hyperrhs
!
!-----------------------------------------------------------------
    subroutine hyperrhs(u_dot, u, v, w, dxu, dyu, dzu, dxv, dyv, dzv, &
         imask, par)

      implicit none

      type(gridfunction), dimension(NU)   :: dxu, dyu, dzu
      type(gridfunction), dimension(NV)   :: dxv, dyv, dzv
      type(gridfunction), dimension(NU)   :: u, u_dot
      type(gridfunction), dimension(NW)   :: w
      type(gridfunction), dimension(NV)   :: v
      CCTK_REAL, dimension(:)             :: par
      CCTK_INT, dimension(:,:,:)          :: imask

      CCTK_REAL, dimension(NU)            :: u_dot_pt, u_pt, dxu_pt, dyu_pt, dzu_pt
      CCTK_REAL, dimension(NV)            :: dxv_pt, dyv_pt, dzv_pt, v_pt
      CCTK_REAL, dimension(NW)            :: w_pt
      CCTK_INT                            :: ii,jj,kk, nx, ny, nz, m
      CCTK_INT                            :: use_mask
      CCTK_REAL                           :: err, myl2norm3d
      CCTK_REAL, pointer, dimension(:,:,:):: mask, x, y, z

      logical, parameter                  :: ltrace = .false.
      CCTK_INT                            :: myid, proc_return_myid
      
      mask => w(H_MASK)%d
      x    => w(H_XPHYS)%d
      y    => w(H_YPHYS)%d
      z    => w(H_ZPHYS)%d


      nx = nint(par(P_NX))
      ny = nint(par(P_NY))
      nz = nint(par(P_NZ))

      use_mask = nint(par(P_USE_MASK))

      !---------------------------------------------------------------------
      ! We first calculate the auxiliary variables.
      !---------------------------------------------------------------------
      if (ltrace) then
         myid = proc_return_myid()
         write(0,*)myid,'] >>> HYPERRHS: calling hyperauxvars'
      end if

      call hyperauxvars(v, w, u, par)      
      
      !---------------------------------------------------------------------
      ! Check to see if derivatives are stored in 3-D arrays.
      !---------------------------------------------------------------------
      
      if (P_POINT_WISE_DERIVATIVES .eq. 0 .and. P_HYPER_DERIVS .eq. 1) then
         if (ltrace) then
            write(0,*)myid,'] >>> HYPERRHS: calling hyperderivs'
         end if
         call hyperderivs(dxu,   dyu,  dzu,  dxv,  dyv,  dzv, &
              u,     v,    w,    imask,      par)
      end if
      
       if (ltrace) write(0,*)myid,'] >>> HYPERRHS: not sure what this does...'
       if (ltrace) write(0,*)myid,'] >>> HYPERRHS: Apparently doing things pointwise'
 
       if (P_HYPER_RHS1 .eq. 1) then
       do kk = 1, nz
          do jj = 1, ny
             do ii = 1, nx
 
 
                !-----------------------------------------------------------
                ! Check to see if we are in an excision region
                !-----------------------------------------------------------
                if (use_mask .eq. 1 .and. mask(ii,jj,kk) < 0.0) then
                   do m=1, NU
                      u_dot(m)%d(ii,jj,kk) = 0.0
                   end do
                else
 
                   do m = 1, NU
                      u_pt(m) = u(m)%d(ii,jj,kk)
                   end do
                   do m = 1, P_NV_REAL_SIZE
                      v_pt(m) = v(m)%d(ii,jj,kk)
                   end do
                   do m = 1, NW
                      w_pt(m) = w(m)%d(ii,jj,kk)
                   end do
 
                   if (P_POINT_WISE_DERIVATIVES .eq. 1) then
                      call hyperderivs_pt(dxu_pt,   dyu_pt,  dzu_pt,        &
                           dxv_pt,   dyv_pt,  dzv_pt,        &
                           u,   v,   w_pt,    imask(ii,jj,kk), & 
                           ii,  jj,  kk,      par)
 
!#if 0
!                     err = 1.0d-12
!                     do m = 1, NU
!                        if (abs(dxu_pt(m) - dxu(m)%d(ii,jj,kk)) > err) then
!                           write(0,*)'error in x deriv ', &
!                                dxu_pt(m),dxu(m)%d(ii,jj,kk)
!                        end if
!                        if (abs(dyu_pt(m) - dyu(m)%d(ii,jj,kk)) > err) then
!                           write(0,*)'error in y deriv ', &
!                                dyu_pt(m),dyu(m)%d(ii,jj,kk)
!                        end if
!                        if (abs(dzu_pt(m) - dzu(m)%d(ii,jj,kk)) > err) then
!                           write(0,*)'error in z deriv ', &
!                                dzu_pt(m),dzu(m)%d(ii,jj,kk)
!                        end if
!
!                     end do
!#endif
                   else
                      do m = 1, NU
                         u_pt(m)   = u(m)%d(ii,jj,kk)
 
                         if (u(m)%take_dx == 1) then
                            dxu_pt(m) = dxu(m)%d(ii,jj,kk)
                         else
                            dxu_pt(m) = 0.0
                         end if
 
                         if (u(m)%take_dy == 1) then
                            dyu_pt(m) = dyu(m)%d(ii,jj,kk)
                         else
                            dyu_pt(m) = 0.0
                         end if
 
                         if (u(m)%take_dz == 1) then
                            dzu_pt(m) = dzu(m)%d(ii,jj,kk)
                         else
                            dzu_pt(m) = 0.0
                         end if
                      end do

                      do m = 1, P_NV_REAL_SIZE
                          if (v(m)%take_dx == 1) then
                            dxv_pt(m) = dxv(m)%d(ii,jj,kk)
                         else
                            dxv_pt(m) = 0.0
                         end if
                         if (v(m)%take_dy == 1) then
                            dyv_pt(m) = dyv(m)%d(ii,jj,kk)
                         else
                            dyv_pt(m) = 0.0
                         end if
 
                         if (v(m)%take_dz == 1) then
                            dzv_pt(m) = dzv(m)%d(ii,jj,kk)
                         else
                            dzv_pt(m) = 0.0
                         end if
                      end do
                   end if
                   
      if (ltrace) write(0,*)myid,'] >>> HYPERRHS: calcrhs on pt:',ii,jj,kk
                   call calcrhs(u_dot_pt, u_pt, v_pt, dxu_pt, dyu_pt, dzu_pt,  &
                        dxv_pt, dyv_pt, dzv_pt, w_pt, par)
 
                   do m=1, NU
                      u_dot(m)%d(ii,jj,kk) = u_dot_pt(m)
                   end do
 
                end if
 
             end do
          end do
       end do
       end if


       if (P_HYPER_RHS2 .eq. 1) then
          call calcrhs2(u_dot, u, v, w, dxu, dyu, dzu, dxv, dyv, dzv, &
               imask, par)
       end if

    end subroutine hyperrhs
end MODULE HYPER_RHS
