!-------------------------------------------------------------------------
!
!  $Id: hyperrk.f90,v 1.2 2006/09/26 02:31:43 inaki Exp $
!
!-------------------------------------------------------------------------

#include "cctk.h"

subroutine HyperEvolveRK(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
     urk2, urk3, par, iter)
  use params
  use GF
  implicit none

  CCTK_INT                         :: iter
  type(gridfunction), dimension(NU):: u0,u2, &
                                      urk1,urk2,urk3,&
                                      dxu,dyu,dzu

  type(gridfunction), dimension(NV):: v, dxv, dyv, dzv
  type(gridfunction), dimension(NW):: w
  CCTK_REAL, dimension(NPAR)           :: par

  ! local vars
  CCTK_INT  :: rk_type

  interface
    subroutine RK3(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
      use params
      use UTILEQS
      use GF
      use HYPER_RHS
      use HYPER_DISSIPATION
      use HYPER_BOUNDARY
      use MOD_MASK
      use hypercoords
      implicit none
      CCTK_INT                         :: iter
      type(gridfunction), dimension(NU):: u0,u2, urk1,urk2,urk3,&
                                          dxu,dyu,dzu
      type(gridfunction), dimension(NV):: v, dxv, dyv, dzv
      type(gridfunction), dimension(NW):: w
      CCTK_REAL, dimension(NPAR)           :: par
    end subroutine RK3
    subroutine TVDRK3(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
      use params
      use UTILEQS
      use GF
      use HYPER_RHS
      use HYPER_DISSIPATION
      use HYPER_BOUNDARY
      use MOD_MASK
      use hypercoords
      implicit none
      CCTK_INT                         :: iter
      type(gridfunction), dimension(NU):: u0,u2, urk1,urk2,urk3,&
                                          dxu,dyu,dzu
      type(gridfunction), dimension(NV):: v, dxv, dyv, dzv
      type(gridfunction), dimension(NW):: w
      CCTK_REAL, dimension(NPAR)           :: par
    end subroutine TVDRK3
    subroutine TVDRK3_FV(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
      use params
      use UTILEQS
      use GF
      use HYPER_RHS
      use HYPER_DISSIPATION
      use HYPER_BOUNDARY
      use MOD_MASK
      use hypercoords
      implicit none
      CCTK_INT                         :: iter
      type(gridfunction), dimension(NU):: u0,u2, urk1,urk2,urk3,&
                                          dxu,dyu,dzu
      type(gridfunction), dimension(NV):: v, dxv, dyv, dzv
      type(gridfunction), dimension(NW):: w
      CCTK_REAL, dimension(NPAR)           :: par
    end subroutine TVDRK3_FV

  end interface

  rk_type = nint(par(P_RUNGE_KUTTA_TYPE))

  if (rk_type .eq. 1) then
    call  RK3(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
  else if (rk_type .eq. 2) then
    call  TVDRK3(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
  else if (rk_type .eq. 3) then
    call  TVDRK3_FV(u0, u2, v, w, dxu,dyu,dzu, dxv, dyv, dzv, urk1, &
         urk2, urk3, par, iter)
  else
    write(0,*)'HyperEvolveRK:  unknown RK type'
  end if
 

  return
end subroutine HyperEvolveRK
