#include "cctk.h"

!$Id: auxvars.f90,v 1.2 2005/11/14 19:56:06 carlos Exp $

module auxvars
use params

implicit none
CONTAINS

!--------------------------------------------------------------------
!
! Define all the quantities that need discrete derivatives but 
! are not the gridfunctions that are evolved (combinations of them with 
! background quantities). In the trivial case of the wave equation we will
! use this variables for the fluxes, which are the variables themselves
! with a opposite sign
!--------------------------------------------------------------------
   subroutine define_aux_vars(v, w, u, par)
   use GF
   
   type(gridfunction), dimension(NV) :: v
   type(gridfunction), dimension(NU) :: u
   type(gridfunction), dimension(NW) :: w
   CCTK_REAL, dimension(:)           :: par

   !! local vars
   CCTK_INT                          :: i, j, k, m, nx, ny, nz
   CCTK_REAL, dimension(NU)          :: upt
   CCTK_REAL, dimension(NV)          :: vpt
   CCTK_REAL, dimension(3)           :: xx
   logical, parameter                :: ltrace2 = .false.
   CCTK_REAL, dimension(:,:,:),pointer :: x, y, z
   target :: w

   x => w(H_XPHYS)%d
   y => w(H_YPHYS)%d
   z => w(H_ZPHYS)%d

   if (ltrace2) write(0,*)'DEFINE_AUX_VARS: begin'

   nx = nint(par(P_NX))
   ny = nint(par(P_NY))
   nz = nint(par(P_NZ))

   if (ltrace2) write(0,*)'DEFINE_AUX_VARS: nx,ny,nz',nx,ny,nz


   do k = 1, nz
      do j = 1, ny
         do i = 1, nx
            xx(1) = x(i,j,k)
            xx(2) = y(i,j,k)
            xx(3) = z(i,j,k)
            
            do m = 1, NU
               upt(m) = u(m)%d(i,j,k)
            end do

            do m = 1, NU
               vpt(m) = v(m)%d(i,j,k)
            end do
            
            !print *,'calling aux_var_pt'
            call aux_var_pt(upt,xx,par,vpt)

            do m = 1, NU
               v(m)%d(i,j,k) = vpt(m)
            end do            

         end do
      end do
   end do
  
    return
 end subroutine define_aux_vars

!--------------------------------------------------------------
!
!
!
!--------------------------------------------------------------
   subroutine aux_var_pt(upt,xx,par,vpt)
      implicit none
      CCTK_REAL, dimension(NU),intent(in)    :: upt
      CCTK_REAL, dimension(NV),intent(out)   :: vpt
      CCTK_REAL, dimension(3),intent(in)     :: xx
      CCTK_REAL, dimension(NPAR),intent(in)  :: par
      ! local vars
      
      vpt(H_FDPHI)   = upt(H_D4PHI)
      vpt(H_FxD4PHI) = upt(H_D1PHI)
      vpt(H_FyD4PHI) = upt(H_D2PHI)
      vpt(H_FzD4PHI) = upt(H_D3PHI)

     return
   end subroutine aux_var_pt

  !--------------------------------------------------------------------
  !
  !
  !
  !--------------------------------------------------------------------

  subroutine SolvePrimVars(v, w, u, par)
    use GF

    type(gridfunction), dimension(NV) :: v
    type(gridfunction), dimension(NU) :: u
    type(gridfunction), dimension(NW) :: w
    CCTK_REAL, dimension(:)           :: par

  end subroutine SolvePrimVars


!----------------------------------------------------------------------
!  This routine is empty
!----------------------------------------------------------------------
   !-----------------------------------------------------------------
   !
   !-----------------------------------------------------------------
    subroutine calcrhs2(u_dot, u, v, w, dxu, dyu, dzu, dxv, dyv, dzv, &
         imask, par)

      use GF
      implicit none

      type(gridfunction), dimension(NU)   :: dxu, dyu, dzu
      type(gridfunction), dimension(NV)   :: dxv, dyv, dzv
      type(gridfunction), dimension(NU)   :: u, u_dot
      type(gridfunction), dimension(NW)   :: w
      type(gridfunction), dimension(NV)   :: v
      CCTK_REAL, dimension(:)             :: par
      CCTK_INT, dimension(:,:,:)          :: imask

      return
    end subroutine calcrhs2


   
 end module auxvars
