cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_pstcompute:                                                          cc
cc            Compute post-cgc correction.                                    cc
cc            Transferred from the next coarser level should                  cc
cc            be u(2h) residing in storage at u_st1                           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_pstcompute(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer       nx,  ny,  nz
      integer       nxc, nyc, nzc
      integer       i, u_ptr, u_st1_ptr, u_st2_ptr
      real(kind=8)  mix, max, miy, may, miz, maz, time

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      if (ltrace) then
         write(*,*) 'grid_pstcompute: grid: ',gi
      end if

      call load_pointers(gi)

      nx  = gr_nx(gi)
      ny  = gr_ny(gi)
      nz  = gr_nz(gi)

      time= gr_t(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )

      mix = gr_minx(gi)
      max = gr_maxx(gi)
      miy = gr_miny(gi)
      may = gr_maxy(gi)
      miz = gr_minz(gi)
      maz = gr_maxz(gi)

         !            h       h    h    2h     2h  h
         !           u    := u  + I  ( u   -  I   u  )
         !                         2h          h
         !
         !                       2h
         !          u_st1   <== u       (from grid_psttranfer)
         !

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            u_ptr     = gfunc_pointer(i,  gi)
            u_st1_ptr = gfunc_pointer(i+1,gi)
            u_st2_ptr = gfunc_pointer(i+2,gi)
            !
            call restrict(q(u_st2_ptr), q(u_ptr),nxc,nyc,nzc)
            if (vtrace) call field_out3d(q(u_st2_ptr),time,'PST_u_2H',
     *                    mix,max,miy,may,miz,maz, nxc,nyc,nzc,myid)
            !
            call mat_mat_sub3d( q(u_st2_ptr), q(u_st1_ptr),q(u_st2_ptr),
     *                                                    nxc,nyc,nzc)
            if (vtrace) call field_out3d(q(u_st2_ptr),time,'PSTcorr2H',
     *                     mix,max,miy,may,miz,maz, nxc,nyc,nzc,myid)
            !
            ! Interpolate correction to fine grid:
            !
            call prolong( q(u_st1_ptr), q(u_st2_ptr), nxc,nyc,nzc)
            if (vtrace) call field_out3d(q(u_st1_ptr),time,'PSTcorrH',
     *                        mix,max,miy,may,miz,maz, nx,ny,nz,myid)
            !
            ! Add correction to field:
            !
            if (vtrace) call field_out3d(q(u_ptr),time,
     *            'PST_u_old', mix,max,miy,may,miz,maz, nx,ny,nz,myid)
             call mat_mat_add3d(q(u_ptr),q(u_ptr),q(u_st1_ptr),nx,ny,nz)
            if (vtrace) call field_out3d(q(u_ptr),time,
     *            'PST_u_new', mix,max,miy,may,miz,maz, nx,ny,nz,myid)
            !
         end if
      end do

      if (ltrace) then
         write(*,*) 'grid_pstcompute: Finished.'
      end if

      return
      end      ! END: grid_pstcompute

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_psttransfer_master:                                                  cc
cc                    This routine owns the parent grid.                      cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_psttransfer_master(child,parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i
      integer     nxt, nyt, nzt, length, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc, owner
      integer       u_ptr, u_st1_ptr, u_st2_ptr
      real(kind=8)  mix, max, miy, may, miz, maz

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      call load_pointers(parent)

      owner = gr_own(child)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent)
      nzp = gr_nz(parent)
      
      !
      ! Dimensions of child:
      !
      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)
      
      mix = gr_minx(child)
      max = gr_maxx(child)
      miy = gr_miny(child)
      may = gr_maxy(child)
      miz = gr_minz(child)
      maz = gr_maxz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)
      
      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))

      ! 
      ! Dimensions of part to be copied:
      !
      nxt    = max_ip - min_ip + 1
      nyt    = max_jp - min_jp + 1
      nzt    = max_kp - min_kp + 1
      
      length = nxt * nyt * nzt
      
      if (length.gt.nxp*nyp*nzp) then 
         write(*,*) 'Problem of space in grid_psttransfer_master'
         call my_exit('Problem of space in grid_psttransfer_master')
      end if

      if (ltrace) then
       write(*,99)myid,'grid_psttransfer_master: grids:   ',child,parent
       write(*,99)myid,'grid_psttransfer_master: nx/y/z : ',nx, ny, nz
       write(*,99)myid,'grid_psttransfer_master: nx/y/zt: ',nxt,nyt,nzt
       write(*,99)myid,'grid_psttransfer_master: nx/y/zc: ',nxc,nyc,nzc
       write(*,99)myid,'grid_psttransfer_master: nx/y/zp: ',nxp,nyp,nzp
       write(*,99)myid,'grid_psttransfer_master: min_i/j/kp:',min_ip,
     *                                              min_jp,min_kp
       write(*,99)myid,'grid_psttransfer_master: min_i/j/kc:',min_ic,
     *                                              min_jc,min_kc
       write(*,99)myid,'grid_psttransfer_master: min_i/j/kcc:',min_icc,
     *                                              min_jcc, min_kcc
       write(*,99)myid,'grid_psttransfer_master: length:     ',length
      end if

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  parent)
            u_st1_ptr = gfunc_pointer(i+1,parent)
            u_st2_ptr = gfunc_pointer(i+2,parent)
            !
            ! Copy data to tmp first:
            !
            call copy_overlap( q(gr_tmp),
     *                         q(u_ptr),
     *                         min_ip, min_jp, min_kp,
     *                         nxt,nyt,nzt,
     *                         nxp,nyp,nzp )
            if (vtrace)
     *         call field_out3d(q(gr_tmp),1.d0*parent, 'PSTT_u_par',
     *                   0.d0,1.d0,0.d0,1.d0,0.d0,1.d0,nxt,nyt,nzt,myid)
            !
            if (ltrace) then
               write(*,99) myid,'       master: sending correction'
               write(*,99) myid,'       master: to owner:   ', owner
               write(*,99) myid,'       master: length:     ',length
            end if
            call MPI_Send(q(gr_tmp), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_PST_DATA, MPI_COMM_WORLD, ierr)
            !
       end if
      end do

      if (ltrace) then
         write(*,99) myid,'grid_psttransfer_master: Finished.'
      end if

 99   format('[',I3,'] ',A,3I7)

      return
      end      ! END: grid_psttransfer_master

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_psttransfer_slave:                                                   cc
cc                    This routine owns the child grid.                       cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_psttransfer_slave(child,parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i, phi_parent
      integer     nxt, nyt, nzt, length, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc, owner
      integer       u_ptr, u_st1_ptr, u_st2_ptr
      real(kind=8)  mix, max, miy, may, miz, maz

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      call load_pointers(child)

      owner = gr_own(parent)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent)
      nzp = gr_nz(parent)
      
      !
      ! Dimensions of child:
      !
      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)
      
      mix = gr_minx(child)
      max = gr_maxx(child)
      miy = gr_miny(child)
      may = gr_maxy(child)
      miz = gr_minz(child)
      maz = gr_maxz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)
      
      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))

      ! 
      ! Dimensions of part to be copied:
      !
      nxt    = max_ip - min_ip + 1
      nyt    = max_jp - min_jp + 1
      nzt    = max_kp - min_kp + 1
      
      length = nxt * nyt * nzt
      
      if (length.gt.nxc*nyc*nzc) then 
         write(*,*) 'Problem of space in grid_psttransfer_slave'
         call my_exit('Problem of space in grid_psttransfer_slave')
      end if

      if (ltrace) then
       write(*,99)myid,'grid_psttransfer_slave: grids:    ',child,parent
       write(*,99)myid,'grid_psttransfer_slave: nx/y/z :  ',nx, ny, nz
       write(*,99)myid,'grid_psttransfer_slave: nx/y/zt:  ',nxt,nyt,nzt
       write(*,99)myid,'grid_psttransfer_slave: nx/y/zc:  ',nxc,nyc,nzc
       write(*,99)myid,'grid_psttransfer_slave: nx/y/zp:  ',nxp,nyp,nzp
       write(*,99)myid,'grid_psttransfer_slave: min_i/j/kc:',min_ic,
     *                                              min_jc,  min_kc
       write(*,99)myid,'grid_psttransfer_slave: min_i/j/kcc:',min_icc,
     *                                              min_jcc, min_kcc
      end if

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  child)
            u_st1_ptr = gfunc_pointer(i+1,child)
            u_st2_ptr = gfunc_pointer(i+2,child)
            if (ltrace) then
               write(*,99) myid,'       slave: receiving correction'
               write(*,99) myid,'       slave: from owner: ', owner
               write(*,99) myid,'       slave: length:     ',length
            end if
            call MPI_Recv(q(gr_tmp), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_PST_DATA, MPI_COMM_WORLD, status,ierr)
  
            !
            ! Transfer into appropriate place in phi_rk1:
            !
            call copy_toplace( q(u_st1_ptr),
     *                         q(gr_tmp),
     *                         min_icc, min_jcc, min_kcc,
     *                         nx, ny, nz,
     *                         nxt,nyt,nzt )
            !
            if (vtrace)
     *         call field_out3d(q(u_st1_ptr),0.d0, 'PSTT_u_2H',
     *                    mix,max,miy,may,miz,maz,nx,ny,nz,myid)
            !
         end if
      end do


      if (ltrace) then
         write(*,99) myid,'grid_psttransfer_slave: Finished.'
      end if

 99   format('[',I3,'] ',A,3I7)

      return
      end      ! END: grid_psttransfer_slave

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_psttransfer:                                                         cc
cc                    This routine gets called for the different              cc
cc                    pairings of child and parent which overlap.             cc
cc                    The goal is to transfer down to the child               cc
cc                    the field(s) u_2H. Once all such pairings finish,       cc
cc                    then grid_pstcompute() processes the complete u_2H      cc
cc                    field to correct u_H.                                   cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_psttransfer(child,parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i, phi_parent
      integer     nxt, nyt, nzt, length, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc
      integer       u_ptr, u_st1_ptr, u_st2_ptr
      real(kind=8)  mix, max, miy, may, miz, maz

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      call load_pointers(parent)
      !phi_parent = gr_phi

      call load_pointers(child)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent)
      nzp = gr_nz(parent)
      
      !
      ! Dimensions of child:
      !
      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)
      
      mix = gr_minx(child)
      max = gr_maxx(child)
      miy = gr_miny(child)
      may = gr_maxy(child)
      miz = gr_minz(child)
      maz = gr_maxz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)
      
      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))

      ! 
      ! Dimensions of part to be copied:
      !
      nxt    = max_ip - min_ip + 1
      nyt    = max_jp - min_jp + 1
      nzt    = max_kp - min_kp + 1
      
      length = nxt * nyt * nzt
      
      if (length.gt.nxc*nyc*nzc) then 
         write(*,*) 'Problem of space in grid_psttransfer'
         call my_exit('Problem of space in grid_psttransfer')
      end if

      if (ltrace) then
         write(*,*)'grid_psttransfer: grids:      ',child, parent
         write(*,*)'grid_psttransfer: nx/y/z :    ',nx, ny, nz
         write(*,*)'grid_psttransfer: nx/y/zt:    ',nxt,nyt,nzt
         write(*,*)'grid_psttransfer: nx/y/zc:    ',nxc,nyc,nzc
         write(*,*)'grid_psttransfer: nx/y/zp:    ',nxp,nyp,nzp
         write(*,*)'grid_psttransfer: min_i/j/kp:', min_ip,
     *                                              min_jp,min_kp
         write(*,*)'grid_psttransfer: min_i/j/kc: ',min_ic,min_jc,min_kc
         write(*,*)'grid_psttransfer: min_i/j/kcc:',min_icc,min_jcc,
     *                                              min_kcc
      end if

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  child)
            u_st1_ptr = gfunc_pointer(i+1,child)
            u_st2_ptr = gfunc_pointer(i+2,child)
            !
            ! Copy data to tmp first:
            !
            call copy_overlap( q(gr_tmp),
     *                         q(gfunc_pointer(i,parent)),
     *                         min_ip, min_jp, min_kp,
     *                         nxt,nyt,nzt,
     *                         nxp,nyp,nzp )
            if (vtrace)
     *      call field_out3d(q(gr_tmp),1.d0*parent, 'PSTT_u_par',
     *                   0.d0,1.d0,0.d0,1.d0,0.d0,1.d0,nxt,nyt,nzt,myid)
            !
            ! Transfer into appropriate place in phi_rk1:
            !
            call copy_toplace(
     *                       q(u_st1_ptr),
     *                       q(gr_tmp),
     *                       min_icc, min_jcc, min_kcc,
     *                       nx, ny, nz,
     *                       nxt,nyt,nzt )

            if (vtrace)
     *      call field_out3d(q(u_st1_ptr),0.d0, 'PSTT_u_2H',
     *                    mix,max,miy,may,miz,maz,nx,ny,nz,myid)
            !
         end if
      end do


      if (ltrace) then
         write(*,*) 'grid_psttransfer: Finished.'
      end if

      return
      end      ! END: grid_psttransfer

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_cgctransfer_slave:                                                   cc
cc                 We own the child but not the parent                        cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_cgctransfer_slave(child, parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxs,nys,nzs
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc
      integer       u_ptr, u_st1_ptr, u_st2_ptr, u_rhs_ptr
      integer     length, owner
      real(kind=8)  mixc, maxc, miyc, mayc, mizc, mazc
      real(kind=8)  mixp, maxp, miyp, mayp, mizp, mazp

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      if (ltrace) then
         write(*,*) 'grid_cgctransfer_slave:child/parent:',child,parent
      end if

      owner = gr_own(parent)

      call load_pointers(child)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent) 
      nzp = gr_nz(parent)

      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      mixc = gr_minx(child)
      maxc = gr_maxx(child)
      miyc = gr_miny(child)
      mayc = gr_maxy(child)
      mizc = gr_minz(child)
      mazc = gr_maxz(child)

      mixp = gr_minx(parent)
      maxp = gr_maxx(parent)
      miyp = gr_miny(parent)
      mayp = gr_maxy(parent)
      mizp = gr_minz(parent)
      mazp = gr_maxz(parent)

      !
      ! Find indices which bound intersection region:
      !
      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)

      !
      ! Overlap region on the parent
      !    but with resolution of the parent grid:
      !
      nxs = max_ip - min_ip + 1
      nys = max_jp - min_jp + 1
      nzs = max_kp - min_kp + 1

      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))
      !
      length = nx * ny * nz

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  child)
            u_st1_ptr = gfunc_pointer(i+1,child)
            u_st2_ptr = gfunc_pointer(i+2,child)
            !
            ! Copy data from child to tmp first:
            !
            call copy_overlap( q(gr_tmp),
     *                         q(u_st1_ptr),
     *                         min_icc, min_jcc, min_kcc,
     *                         nxs,nys,nzs,
     *                         nx ,ny ,nz  )
            call copy_overlap( q(gr_flag),
     *                         q(u_st2_ptr),
     *                         min_icc, min_jcc, min_kcc,
     *                         nxs,nys,nzs,
     *                         nx ,ny ,nz  )
            !
            ! Send to owner of parent:
            !
            if (ltrace) then
               write(*,99) myid,' slave:  sending correction'
               write(*,99) myid,' slave:   to owner: ', owner
            end if
            !
            call MPI_Send(q(gr_tmp), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_CGC_DATA, MPI_COMM_WORLD, ierr)
            call MPI_Send(q(gr_flag), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_CGC_DATA, MPI_COMM_WORLD, ierr)
            !
         end if
      end do

 99   format('[',I3,'] ',A,3I7)
      if (ltrace) then
         write(*,*) 'grid_cgctransfer_slave: Finished.'
      end if

      return
      end      ! END: grid_cgctransfer_slave

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_cgctransfer_master:                                                  cc
cc                               We own parent.                               cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_cgctransfer_master(child, parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxs,nys,nzs
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc
      integer       u_ptr, u_st1_ptr, u_st2_ptr, u_rhs_ptr
      integer     length, owner
      real(kind=8)  mixc, maxc, miyc, mayc, mizc, mazc
      real(kind=8)  mixp, maxp, miyp, mayp, mizp, mazp


      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      if (ltrace) then
         write(*,*)'grid_cgctransfer_master:child/parent:',child,parent
      end if

      owner = gr_own(child)

      call load_pointers(parent)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent) 
      nzp = gr_nz(parent)

      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      mixc = gr_minx(child)
      maxc = gr_maxx(child)
      miyc = gr_miny(child)
      mayc = gr_maxy(child)
      mizc = gr_minz(child)
      mazc = gr_maxz(child)

      mixp = gr_minx(parent)
      maxp = gr_maxx(parent)
      miyp = gr_miny(parent)
      mayp = gr_maxy(parent)
      mizp = gr_minz(parent)
      mazp = gr_maxz(parent)

      !
      ! Find indices which bound intersection region:
      !
      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)

      !
      ! Overlap region on the parent
      !    but with resolution of the parent grid:
      !
      nxs = max_ip - min_ip + 1
      nys = max_jp - min_jp + 1
      nzs = max_kp - min_kp + 1

      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))

      length = nx * ny * nz
      if (length .gt. nxp*nyp*nzp) then
         write(*,*) 'grid_cgctransfer_master: Not enough mem'
         call my_exit('grid_cgctransfer_master: Not enough mem')
      end if

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  parent)
            u_st1_ptr = gfunc_pointer(i+1,parent)
            u_st2_ptr = gfunc_pointer(i+2,parent)
            u_rhs_ptr = gfunc_pointer(i+3,parent)
            !
            if (ltrace) then
               write(*,99) myid,' master: receiving correction'
               write(*,99) myid,' master: from owner: ', owner
            end if
            !
            call MPI_Recv(q(gr_tmp), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_CGC_DATA, MPI_COMM_WORLD, status,ierr)
            call MPI_Recv(q(gr_flag), length, MPI_DOUBLE_PRECISION,
     *                 owner, TAG_CGC_DATA, MPI_COMM_WORLD, status,ierr)
            !
            ! Transfer into appropriate place on parent:
            !
            if (vtrace)
     *         call field_out3d(q(u_ptr),0.d0, 'CGCT_old_par',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            call copy_toplace( q(u_ptr),
     *                         q(gr_tmp),
     *                         min_ip, min_jp, min_kp,
     *                         nxp,nyp,nzp,
     *                         nxs,nys,nzs )
            if (vtrace)
     *         call field_out3d(q(u_ptr),0.d0, 'CGCT_new_par',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            !
            if (vtrace)
     *         call field_out3d(q(u_rhs_ptr),0.d0, 'CGCT_old_rho',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            call copy_toplace( q(u_rhs_ptr),
     *                         q(gr_flag),
     *                         min_ip, min_jp, min_kp,
     *                         nxp,nyp,nzp,
     *                         nxs,nys,nzs )
            if (vtrace)
     *         call field_out3d(q(u_rhs_ptr),0.d0, 'CGCT_new_rho',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            !
         end if
      end do

      if (ltrace) then
         write(*,*) 'grid_cgctransfer_master: Finished.'
 99      format('[',I3,'] ',A,3I7)
      end if

      return
      end      ! END: grid_cgctransfer_master

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_cgctransfer:                                                         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_cgctransfer(child, parent)
      implicit      none
      integer       child, parent
      include      'grid.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      integer     nxs,nys,nzs
      integer     nxp, nyp, nzp, nxc, nyc, nzc, i, nx,ny,nz
      integer     min_ic,max_ic, min_jc,max_jc, min_kc,max_kc
      integer     min_ip,max_ip, min_jp,max_jp, min_kp,max_kp
      integer     min_icc, min_jcc, min_kcc
      integer       u_ptr, u_st1_ptr, u_st2_ptr, u_rhs_ptr
      integer     length
      real(kind=8)  mixc, maxc, miyc, mayc, mizc, mazc
      real(kind=8)  mixp, maxp, miyp, mayp, mizp, mazp


      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      if (ltrace) then
         write(*,*) 'grid_cgctransfer: child, parent: ',child, parent
      end if

      !call load_pointers(parent)
      !parent_phi = gr_phi
      !parent_rho = gr_rho
      call load_pointers(child)

      nxp = gr_nx(parent)
      nyp = gr_ny(parent) 
      nzp = gr_nz(parent)

      nxc = gr_nx(child)
      nyc = gr_ny(child)
      nzc = gr_nz(child)

      !
      ! Dimensions of virtual coarse grid (the u(2h) grid);
      !
      nx  = NINT( 0.5d0*(nxc+1) )
      ny  = NINT( 0.5d0*(nyc+1) )
      nz  = NINT( 0.5d0*(nzc+1) )

      mixc = gr_minx(child)
      maxc = gr_maxx(child)
      miyc = gr_miny(child)
      mayc = gr_maxy(child)
      mizc = gr_minz(child)
      mazc = gr_maxz(child)

      mixp = gr_minx(parent)
      maxp = gr_maxx(parent)
      miyp = gr_miny(parent)
      mayp = gr_maxy(parent)
      mizp = gr_minz(parent)
      mazp = gr_maxz(parent)

      !
      ! Find indices which bound intersection region:
      !
      call grid_find_intersection(child, parent, length,
     *            min_ic, max_ic, min_jc, max_jc, min_kc, max_kc)
      call grid_find_intersection(parent, child, length,
     *            min_ip, max_ip, min_jp, max_jp, min_kp, max_kp)

      !
      ! Overlap region on the parent
      !    but with resolution of the parent grid:
      !
      nxs = max_ip - min_ip + 1
      nys = max_jp - min_jp + 1
      nzs = max_kp - min_kp + 1

      ! 
      ! We're creating a 2h field on this grid, so we need
      ! the location of the intersection of the parent with
      ! this 2h child grid:
      !        (NB: interesction should be odd because
      !             child has to intersect parent at parent's grid points)
      ! 
      min_icc = NINT(0.5d0*(min_ic+1))
      min_jcc = NINT(0.5d0*(min_jc+1))
      min_kcc = NINT(0.5d0*(min_kc+1))

      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  child)
            u_st1_ptr = gfunc_pointer(i+1,child)
            u_st2_ptr = gfunc_pointer(i+2,child)
            u_rhs_ptr = gfunc_pointer(i+3,child)
            !
            !
            ! Copy data from child to tmp first:
            !
            call copy_overlap( q(gr_tmp),
     *                         q(u_st1_ptr),
     *                         min_icc, min_jcc, min_kcc,
     *                         nxs,nys,nzs,
     *                         nx ,ny ,nz  )
            call copy_overlap( q(gr_flag),
     *                         q(u_st2_ptr),
     *                         min_icc, min_jcc, min_kcc,
     *                         nxs,nys,nzs,
     *                         nx ,ny ,nz  )
            !
            ! Transfer into appropriate place on parent:
            !
            !
            u_ptr     = gfunc_pointer(i,  parent)
            u_st1_ptr = gfunc_pointer(i+1,parent)
            u_st2_ptr = gfunc_pointer(i+2,parent)
            u_rhs_ptr = gfunc_pointer(i+3,parent)
            if (vtrace)
     *         call field_out3d(q(u_ptr),0.d0, 'CGCT_old_par',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            call copy_toplace( q(u_ptr),
     *                         q(gr_tmp),
     *                         min_ip, min_jp, min_kp,
     *                         nxp,nyp,nzp,
     *                         nxs,nys,nzs )
            if (vtrace)
     *         call field_out3d(q(u_ptr),0.d0, 'CGCT_new_par',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            !
            if (vtrace)
     *         call field_out3d(q(u_rhs_ptr),0.d0, 'CGCT_old_rho',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            call copy_toplace( q(u_rhs_ptr),
     *                         q(gr_flag),
     *                         min_ip, min_jp, min_kp,
     *                         nxp,nyp,nzp,
     *                         nxs,nys,nzs )
            if (vtrace)
     *         call field_out3d(q(u_rhs_ptr),0.d0, 'CGCT_new_rho',
     *                   mixp,maxp,miyp,mayp,mizp,mazp,nxp,nyp,nzp,myid)
            !
         end if
      end do

      if (ltrace) then
         write(*,*) 'grid_cgctransfer: Finished.'
      end if

      return
      end      ! END: grid_cgctransfer

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_cgccompute:                                                          cc
cc            Compute coarse grid correction.                                 cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_cgccompute(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      include      'glob.inc'
      include      'mpif.h'
      include      'mpi_stuff.inc'
      real(kind=8)  hf, hc
      integer       nx,  ny,  nz
      integer       nxc, nyc, nzc
      real(kind=8)  mix, max, miy, may, miz, maz, time
      integer       i, u_ptr, u_st1_ptr, u_st2_ptr, u_rhs_ptr
      character*4   tmpchar

      logical     ltrace
      parameter ( ltrace = .false. )
      logical     vtrace
      parameter ( vtrace = .false. )


      if (ltrace) then
         write(*,*) 'grid_cgccompute: grid: ',gi
      end if

      call load_pointers(gi)

      nx  = gr_nx(gi)
      ny  = gr_ny(gi)
      nz  = gr_nz(gi)

      hf  = gr_h(gi)
      hc  = 2.d0*hf

      time= gr_t(gi)

      mix = gr_minx(gi)
      max = gr_maxx(gi)
      miy = gr_miny(gi)
      may = gr_maxy(gi)
      miz = gr_minz(gi)
      maz = gr_maxz(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )

         !
         !  (a)      Compute rhs on (m-1) level:
         !
         !              2h      2h 2h h     2h       h    h h
         !           rhs   :=  L  I  u   + I   (  rhs  - L u  )
         !                          h        h
         !
         !  (b)      Restrict u and transfer to the (m-1) level
         !              
         !            2h         2h  h
         !           u       := I   u
         !                       h

      ! Use the associated storage arrays as temporary memory:
      !
      !                     2h
      ! u_st1       <===   u
      !
      !                       2h
      ! u_st2       <===   rhs
      !
            !
            ! (1) Compute residual of u^h  into the "u_st1" fields
            ! (2) Restrict            u^h    to the "u_st2" fields
            ! (3) Compute lop      of u^2h into the "u_st2" fields
            !
      if (ltrace) write(*,*) 'compute: call grid_get_resid()'
      call grid_get_resid(gi)
      !
      if (ltrace) write(*,*) 'compute: restricting fields & resids'
      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  gi)
            u_st1_ptr = gfunc_pointer(i+1,gi)
            u_st2_ptr = gfunc_pointer(i+2,gi)
            !
            call restrict(  q(u_st2_ptr), q(u_st1_ptr), nxc,nyc,nzc)
            call restrict(  q(u_st1_ptr), q(u_ptr),     nxc,nyc,nzc)
            !
         end if
      end do
      !
      if (ltrace) write(*,*) 'compute: call grid_lop()'
      call grid_lop(gi)
      !
      if (ltrace) write(*,*) 'compute: computing new rhs'
      do i = 1, num_gfuncs
         if ( gfunc_type(i) .eq. GFUNC_INTEGRAL_DEREL ) then
            !
            u_ptr     = gfunc_pointer(i,  gi)
            u_st1_ptr = gfunc_pointer(i+1,gi)
            u_st2_ptr = gfunc_pointer(i+2,gi)
            !
            ! Mask out DD boundaries where data is on a 2h domain
            ! and chr is on a h domain:
            !
            if (ltrace) write(*,*) 'compute: masking'
            call maskLargeh2h(q(u_st2_ptr),q(gr_chr),nxc,nyc,nzc,
     *                                                      bound_width)
            if (ltrace) write(*,*) 'compute: masking'
            call maskLargeh2h(q(u_st1_ptr),q(gr_chr),nxc,nyc,nzc,
     *                                                      bound_width)
            !
            if (vtrace) then
               call int2str(i,tmpchar)
               call field_out3d(q(u_st2_ptr),time,'CGC_rhs_2H'//tmpchar,
     *                       mix,max,miy,may,miz,maz,nxc,nyc,nzc,myid)
               call field_out3d(q(u_st1_ptr),time, 'CGC_u_2H'//tmpchar,
     *                       mix,max,miy,may,miz,maz,nxc,nyc,nzc,myid)
            end if
            !
         end if
      end do

      if (ltrace) then
         write(*,*) 'grid_cgccompute: Finished.'
      end if

      return
      end      ! END: grid_cgcompute
