!Do NOT Edit this file...it has been copied here from /home/matt/had-example/src/hyperCurvedWave
!
!
! grid_ell.f:  These are a collection of project-specific routines
!              which address the elliptic problem. If the project
!              has no elliptic equation to solve, these should just
!              be dummy routines (declared, but no body necessary).
!
!              The idea is that the call sequences to the "normal" 
!              lop(), get_resid(), and relax() routines need to be 
!              hidden from the infrastructure and kept secured in  
!              the project directories.                             
!
!
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_get_resid:                                                           cc
cc            Compute the residual of the current solution and stick          cc
cc            it into the u_st1 arrays.                                       cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_get_resid( gi )
      implicit    none
      integer     gi
      include    'grid.inc'
      integer     nx,  ny,  nz
      integer     nxc, nyc, nzc
      real*8      hg, hc
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index

      call load_pointers(gi)

      if (ltrace) write(*,*) 'grid_get_resid: On grid: ',gi

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )
      hc  = 2.d0*hg

      call get_resid(q(gr_psi_elliptic_st1),
     *               q(gr_psi_elliptic),
     *               q(gr_psi_elliptic_rhs),hg,nx,ny,nz)

      return
      end    ! END: grid_get_resid

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_lop:                                                                 cc
cc            Compute the results of the differential operator associated     cc
cc            with the elliptic problem acting on the set of restricted       cc
cc            fields residing in the u_st1 fields and *add* the result        cc
cc            to the u_st2 arrays.                                            cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_lop( gi )
      implicit    none
      integer     gi
      include    'grid.inc'
      integer     nx,  ny,  nz
      integer     nxc, nyc, nzc
      real*8      hg, hc
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index

      call load_pointers(gi)

      if (ltrace) write(*,*) 'grid_lop:    On grid: ',gi

      nx   = gr_nx(gi)
      ny   = gr_ny(gi)
      nz   = gr_nz(gi)
      hg   = gr_h(gi)

      nxc = NINT( 0.5d0*(nx+1) )
      nyc = NINT( 0.5d0*(ny+1) )
      nzc = NINT( 0.5d0*(nz+1) )
      hc  = 2.d0*hg

      call lop( q(gr_psi_elliptic_st2),
     *          q(gr_psi_elliptic_st1),hc,nxc,nyc,nzc)

      return
      end    ! END: grid_lop

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_ellrhs:                                                              cc
cc               Compute the RHS for the elliptic solve.                      cc
cc               If no elliptic solver in project, just leave empty.          cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_ellrhs( gridnum )
      implicit    none
      integer     gridnum
      include    'grid.inc'
      integer     nx, ny, nz
      real*8      hg
      logical     ltrace
      parameter ( ltrace = .false. )
      integer    i,j,k,index
      integer    indexip1, indexim1
      integer    indexjp1, indexjm1
      integer    indexkp1, indexkm1

      call load_pointers(gridnum)

      return
      end    ! END: grid_ellrhs

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_relax:                                                               cc
cc            Apply relaxation routine to field(s) on a given grid.           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_relax(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      include      'glob.inc'

      real(kind=8)   relax
      external       relax
      real(kind=8)   resid

      logical     ltrace
      parameter ( ltrace  = .false. )


      if (ltrace) then
         write(*,*) 'grid_relax: Relaxing grid: ',gi
      end if

      call load_pointers(gi)

      !resid = relax(q(gr_phi),q(gr_rho),q(gr_chr),
      resid = relax(q(gr_psi_elliptic),q(gr_psi_elliptic_rhs),q(gr_chr),
     *                gr_h(gi),gr_nx(gi),gr_ny(gi),gr_nz(gi))

      !
      ! For output and debugging purposes only:
      !
!     call get_resid(q(gr_phi_x),q(gr_phi),q(gr_rho),
!    *                gr_h(gi),gr_nx(gi),gr_ny(gi),gr_nz(gi))

      if (ltrace) then
         write(*,*) 'grid_relax: Finished with resid: ',resid
      end if

      return
      end      ! END: grid_relax

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_return_resnrm:  Get norm of residual on a given grid (for debugging):cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      double precision function grid_return_resnrm(gi)
      implicit      none
      integer       gi
      include      'grid.inc'
      integer       nx, ny, nz

      real(kind=8)  h
      real(kind=8)  get_resnrm
      external      get_resnrm
      real(kind=8)  grid_return_h
      external      grid_return_h
      logical       grid_is_local
      external      grid_is_local

      logical     ltrace
      parameter ( ltrace = .false. )


      call grid_get_dims(gi,nx,ny,nz)
      h = grid_return_h(gi)

      call load_pointers(gi)

      if (.not. grid_is_local(gi)) then
         write(*,*) 'grid_return_resnrm: Only works on local grids'
      end if

      grid_return_resnrm = get_resnrm(q(gr_psi_elliptic),
     *                          q(gr_psi_elliptic_rhs),h,nx,ny,nz)

      return
      end      ! END: grid_return_resnrm

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  grid_ellpost:                                                             cc
cc               Things to do after an elliptic solve.                        cc
cc               For MHD project, this is where we clean the B field.         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grid_ellpost( gridnum )
      implicit    none
      integer     gridnum
      include    'grid.inc'
      include    'param.inc'
      integer     nx, ny, nz
      real*8      hg
      logical     ltrace
      parameter ( ltrace = .true. )
      integer    i,j,k,index
      integer    indexip1, indexim1
      integer    indexjp1, indexjm1
      integer    indexkp1, indexkm1
      real*8      grad_x_psi, grad_y_psi, grad_z_psi

      call load_pointers(gridnum)

      return
      end    ! END: grid_ellpost

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  get_resid:                                                                cc
cc             Compute:   res = rhs - L u                                     cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine get_resid(res, phi, rhs, h, nx, ny, nz)
      implicit      none
      integer       nx, ny, nz
      real(kind=8)  res(nx,ny,nz), phi(nx,ny,nz), rhs(nx,ny,nz), h


      return
      end      ! END: get_resid

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  get_resnrm:                                                               cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      double precision function get_resnrm(psi,rhs,h,nx,ny,nz)
      implicit      none
      integer       nx, ny, nz
      real(kind=8)  psi(nx,ny,nz), rhs(nx,ny,nz), h
      integer       i,j,k

      get_resnrm = 0.d0
      return
      end       ! END: get_resnrm

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  lop:  Needs to parallel the finite differencing in relax()                cc
cc        NB: Needs to add the calculation to what resides in                 cc
cc            storage already!                                                cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine lop(result,psi,h,nx,ny,nz)
      implicit      none
      integer       nx, ny, nz
      real(kind=8)  result(nx,ny,nz), psi(nx,ny,nz), h

      real(kind=8)  dpsi_dx2, dpsi_dy2, dpsi_dz2
      integer       i, j, k

      logical     ltrace
      parameter ( ltrace = .false. )

      return
      end      ! END: lop

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  relax:                                                                    cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      double precision function relax(psi,rhs,chr,h,nx,ny,nz)
      implicit      none
      integer       nx, ny, nz
      real(kind=8)  psi(nx,ny,nz), rhs(nx,ny,nz), chr(nx,ny,nz), h

      relax = 0.0d0

      return
      end      ! END: relax

