cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  ell_solve:                                                                cc
cc                Top level routine to do an elliptic solve.                  cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine ell_solve( )
      implicit none
      include     'grid_methods.inc'
      include     'grid.inc'
      include     'glob.inc'
      include     'action.inc'
      include     'mpif.h'
      include     'mpi_stuff.inc'
      include     'output.inc'
      include     'param.inc'
      real(kind=8) residnorm
      integer      i, finest, l, cycle
      ! coarsest grid nx/y/z
      integer      nxc, nyc, nzc       
      ! factors of 2 coarsening for each dim
      integer      nfact_x, nfact_y, nfact_z 
      ! return codes from factoring
      integer      xrc, yrc, zrc

      logical      ltrace
      parameter (  ltrace = .true. )
      logical      ltrace2
      parameter (  ltrace2 = .false. )

      !
      ! Else solve by multigrid:
      !
      logical      solve_by_relaxation
      parameter (  solve_by_relaxation  = .false. )

      ! Multigrid constants:
      integer         lmax,          minnxc,           maxnxc
      parameter     ( lmax = 12,     minnxc = 5,       maxnxc = 9  )

      logical      do_nothing
      parameter (  do_nothing           = .false. )

      integer         checkfacts
      external        checkfacts

      if (do_nothing) return

      finest = level_return_finest(Levelp,maxlev)

      if (ltrace2) then
         write(*,*) 'ell_solve: Beginning on level: ',finest
      end if

      !
      ! Compute the RHS for the elliptic problem:
      !
      call level_apply(finest,ELLRHS)

      if (solve_by_relaxation) then
         !
         ! Only relax finest existing level:
         !
         if (ltrace) write(*,*) 'ell_solve: Solve by relaxation:',finest
         do i = 1, maxsweeps
            call level_apply(finest,RELAX)
            call level_apply(finest,SYNCBNDI)
         end do
         if (ltrace) write(*,*) 'ell_solve: Relaxation succeeded',i-1
      else
         !
         ! Working only with coarsest grid
         !
         if (ltrace .and. .false.) then
            write(*,*) '******** parameter dump ************'
            write(*,*) '    number of levels            l =', l
            write(*,*) '    number of v-cycles    nvcycle =', nvcycle
            write(*,*) '    number of pre-sweeps   preswp =', preswp
            write(*,*) '    number of post-sweeps  pstswp =', pstswp
            write(*,*) '************ end dump **************'
         end if
         !-------------------------------------
         ! Factorize Grid Handed-in into Sub-Grids
         !        minnxc,  maxnyc  --- minimum / maximum size of coarse grid
         !        nx0,     ny0     --- fine grid resolution as passed into routine
         !        nfact_x, nfact_y --- number of factors of two in (nx0,ny0)
         !-------------------------------------
         xrc = checkfacts(nx0 - 1,maxnxc,minnxc,nxc,nfact_x)
         yrc = checkfacts(ny0 - 1,maxnxc,minnxc,nyc,nfact_y)
         zrc = checkfacts(nz0 - 1,maxnxc,minnxc,nzc,nfact_z)
         if( xrc .lt. 0  .or.  yrc .lt. 0 .or. zrc .lt. 0) then
            write(*,*) 'solve_fas: Grid of size: ', nx0,ny0,nz0
            write(*,*) nx0-1,' factors to ',nfact_x,' factors of 2'
            write(*,*) ' with remaining: ', nxc
            write(*,*) ny0-1,' factors to ',nfact_y,' factors of 2'
            write(*,*) ' with remaining: ', nyc
            write(*,*) nz0-1,' factors to ',nfact_z,' factors of 2'
            write(*,*) ' with remaining: ', nzc
            write(*,*) 'solve_fas: not factorizable down to'
            write(*,*) 'solve_fas: maximum coarse size: ',maxnxc
            write(*,*) 'solve_fas: Quitting...'
            call my_exit('Cannot factorize coarse grid sufficiently')
         end if
         !
         !     Set number of levels and coarse grid res:
         !
         if (nfact_x. lt. nfact_y .and. nfact_x .lt. nfact_z) then
               l   = nfact_x + 1
               nxc = 1 + nxc
               nyc = 1 + nyc*(2**(nfact_y-nfact_x))
               nzc = 1 + nzc*(2**(nfact_z-nfact_x))
         else if (nfact_y .le. nfact_x .and. nfact_y .lt. nfact_z) then
               l   = nfact_y + 1
               nxc = 1 + nxc*(2**(nfact_x-nfact_y))
               nyc = 1 + nyc
               nzc = 1 + nzc*(2**(nfact_z-nfact_y))
         else 
               l   = nfact_z + 1
               nxc = 1 + nxc*(2**(nfact_x-nfact_z))
               nyc = 1 + nyc*(2**(nfact_y-nfact_z))
               nzc = 1 + nzc
         end if
         !
         !     Check number of levels:
         !
         if ( l .gt. lmax ) then
             write(*,*) 'ell_solve: too many levels ', l,lmax
             call my_exit('Too many levels')
         end if
         !
         if (ltrace .and. .false.) then
           write(*,*)'ell_solve: l:           ',l
           write(*,*)'ell_solve: nx/y/z0:     ',nx0,ny0,nz0
           write(*,*)'ell_solve: nx/y/zc:     ',nxc,nyc,nzc
           write(*,*)'ell_solve: nfact_x/y/z: ',nfact_x,nfact_y,nfact_z
         end if
         !
         !-------------------------------------
         ! Create grid structure
         !-------------------------------------

         !-------------------------------------
         ! Solve the problem with V-Cycling
         !-------------------------------------
         do cycle = 1, nvcycle
            call vcycle(finest)
         end do
         !
      end if

      !
      ! Do any postprocessing necessary:
      !
      call level_apply(finest,ELLPOST)

      return
      end      ! END: ell_solve

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  vcycle:                                                                   cc
cc          Right now, just using AMR hierarchy. At some point,               cc
cc          will need to create appropriate memory structures for             cc
cc          "grids" coarser than the AMR hierarchy's coarsest grid.           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine vcycle(l)
      implicit      none
      integer       l
      include      'glob.inc'
      include      'action.inc'
      include      'grid_methods.inc'
      integer       m, sweep
      real(kind=8)  resnrm

      !
      ! Output coarse grid results with this frequency:
      !
      integer       cg_period
      parameter (   cg_period  = 50 )

      logical      ltrace
      parameter (  ltrace  = .true. )
      logical      ltrace2
      parameter (  ltrace2 = .false. )


      if (ltrace) then
         write(*,*) 'vcycle:                  / / / *******\\ \\ \\'
      end if
      if (ltrace2) then
         write(*,*) 'vcycle:                       l      = ',l
         write(*,*) 'vcycle:                       preswp = ',preswp
         write(*,*) 'vcycle:                       pstswp = ',pstswp
         write(*,*) 'vcycle:'
      end if
      !-----------------------------------
      !    Coarse grid correction
      !-----------------------------------
      do m = l, 1, -1
         !
         ! Pre-CGC relaxation sweeps:
         !
         if (ltrace2) write(*,*) 'vcycle: Presweeping  level: ',m
         do sweep = 1, preswp
            call level_apply(m,RELAX)
            call level_apply(m,SYNCBNDI)
            resnrm = level_return_resnrm(m)
            if (ltrace) call output_resid(m,'Pre-CGC',resnrm)
         end do
         !
         !
         !  (a)      Compute rhs on (m-1) level:
         !
         !              2h      2h 2h h     2h       h    h h
         !           rhs   :=  L  I  u   + I   (  rhs  - L u  )
         !                          h        h
         !
         !  (b)      Restrict u and transfer to the (m-1) level
         !              
         !            2h         2h  h
         !           u       := I   u
         !                       h
         !
         if (ltrace2) write(*,*) 'vcycle: Computing      CGC level: ',m
         call level_apply(m,MG_CGCCOMPUTE)
         call level_apply(m,MG_CGCTRANSFER)
      end do

      !-----------------------------------
      !    Solve coarse grid problem
      !-----------------------------------
      ! 
      if (ltrace2) write(*,*) 'vcycle: Solving coarse grid problem'
      sweep  = 0
      resnrm = 2.d0*ell_epsilon
      do while (resnrm .gt. ell_epsilon .and. sweep .lt. maxsweeps)
            call level_apply(0,RELAX)
            call level_apply(0,SYNCBNDI)
            resnrm = level_return_resnrm(0)
            if (ltrace .and. mod(sweep,cg_period).eq.0)
     *               call output_resid(0,'CG Solve:',resnrm)
            sweep = sweep + 1
      end do
      if (ltrace) write(*,13) 'Coarse grid solve took: ',sweep,
     *                   ' sweeps w/ resnrm= ',resnrm
 13   format(A,I6,a,1p,E8.2)
      if (resnrm .gt. ell_epsilon) then
         write(*,*) 'vcycle: Coarse grid solve failed.'
         !call my_exit('vcycle: Coarse grid solve failed.')
      end if

      !-----------------------------------
      !    Post-CGC fix & relaxation
      !-----------------------------------
      do m = 1, l
         !           Linearly interpolate correction and update
         !           level m unknown.
         !
         !            h       h    h    2h     2h  h
         !           u    := u  + I  ( u   -  I   u  )
         !                         2h          h
         !
         if (ltrace2) write(*,*) 'vcycle: Computing Post-CGC level: ',m
         call level_apply(m,MG_PSTTRANSFER)
         if (ltrace2) write(*,*) 'vcycle: Computing Post-CGC compute'
         call level_apply(m,MG_PSTCOMPUTE)
         if (ltrace2) write(*,*) 'vcycle: Computing Syncing'
         call level_apply(m,SYNCBNDI)

         !
         ! Post-CGC relaxation sweeps:
         !
         if (ltrace2) write(*,*) 'vcycle: Postsweeping level: ',m
         do sweep = 1, pstswp
            call level_apply(m,RELAX)
            call level_apply(m,SYNCBNDI)
            resnrm = level_return_resnrm(m)
            if (ltrace) call output_resid(m,'Post-CGC',resnrm)
         end do
         !
      end do


      if (ltrace) then
         call level_get_resid(l)
         write(*,*) 'vcycle:                  \\ \\ \\ *******/ / / '
      end if

      return
      end      ! END: vcycle

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  level_return_resnrm:                                                      cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      double precision function level_return_resnrm(l)
      implicit      none
      integer       l
      integer       gi, numpoints, totalpts
      integer       nx, ny, nz
      real(kind=8)  norm
      real(kind=8)  grid_return_resnrm
      external      grid_return_resnrm
      integer       level_return_start   , grid_return_sibling
      external      level_return_start   , grid_return_sibling
      logical       grid_return_existence, grid_is_local
      external      grid_return_existence, grid_is_local

      logical       ltrace
      parameter (   ltrace = .false. )


      if (ltrace) write(*,*) 'level_return_resnrm: Level: ',l

      totalpts            = 0
      level_return_resnrm = 0.d0

      gi = level_return_start(l)
 7    if (grid_return_existence(gi)) then
          call grid_get_dims(gi,nx,ny,nz)
          numpoints = nx * ny * nz
          totalpts  = totalpts + numpoints
          if (grid_is_local(gi)) then
             norm      = grid_return_resnrm(gi)
          end if
          if (ltrace) write(*,*) 'level_return_resnrm: norm(gi) =',norm
          level_return_resnrm = level_return_resnrm + numpoints *norm**2
          gi = grid_return_sibling(gi)
          goto 7
      end if

      if (totalpts .ne. 0) then
         level_return_resnrm = sqrt(level_return_resnrm/totalpts)
      end if

      return
      end       ! END: level_return_resnrm

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  level_get_resid:  Computes residual and sticks it into the _st1 fields.   cc
cc                    Used just for being able to output the resid (for now). cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine level_get_resid(l)
      implicit      none
      integer       l
      integer       gi, numpoints, totalpts
      integer       nx, ny, nz
      real(kind=8)  norm
      real(kind=8)  grid_return_resnrm
      external      grid_return_resnrm
      integer       level_return_start   , grid_return_sibling
      external      level_return_start   , grid_return_sibling
      logical       grid_return_existence, grid_is_local
      external      grid_return_existence, grid_is_local

      logical       ltrace
      parameter (   ltrace = .false. )


      if (ltrace) write(*,*) 'level_get_resid:     Level: ',l

      gi = level_return_start(l)
 7    if (grid_return_existence(gi)) then
          if (grid_is_local(gi)) then
             call grid_get_resid(gi)
          end if
          if (ltrace) write(*,*) 'level_get_resid:    gi = ',gi        
          gi = grid_return_sibling(gi)
          goto 7
      end if

      return
      end       ! END: level_get_resid

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  level_bbox_write:    Write to file the clustering information             cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine level_bbox_write(li,b_minx,b_maxx,b_miny,b_maxy,
     *                              b_minz,b_maxz, time, numbox)
      implicit    none
      integer     li,        numbox
      real(kind=8)      time
      integer     b_minx(*), b_maxx(*),
     *            b_miny(*), b_maxy(*),
     *            b_minz(*), b_maxz(*)
      integer     filenum,   i
      character(1) junk
      include     'param.inc'

      logical     ltrace
      parameter ( ltrace = .false. )

      if (numbox.le.0) return

      if (ltrace) write(*,*) 'level_bbox_write: Recording for level:',li

      filenum = 9
      open(filenum,file='.cluster',form='formatted',status='unknown')

      ! Get to the end of the file to append info:
         junk = "Z"
  6      format(A1)
  8      read(filenum,6,end=10) junk
         goto 8 
 10      continue  

      if (ltrace) write(*,*) 'time,gi,numbox: ',time,li,numbox

      write(filenum,99) time, li, numbox, nx0,ny0,nz0
  99  format(E18.9,5I7)

      do i = 1, numbox
         write(filenum,98) b_minx(i),b_miny(i),b_minz(i),
     *                     b_maxx(i),b_maxy(i),b_maxz(i)
         if (ltrace)
     *         write(*,97) b_minx(i),b_maxx(i),
     *                     b_miny(i),b_maxy(i),
     *                     b_minz(i),b_maxz(i)
      end do
  98  format(6I8)
  97  format(2I4,'  ',2I4,'  ',2I4)

      close(filenum)

      return
      end      ! END: level_bbox_write

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  level_bbox_read:     Read from file the clustering information            cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine level_bbox_read(li,b_minx,b_maxx,b_miny,b_maxy,
     *                              b_minz,b_maxz, time, numbox,status)
      implicit    none
      integer     li,        numbox, status
      real(kind=8)      time
      integer     b_minx(*), b_maxx(*),
     *            b_miny(*), b_maxy(*),
     *            b_minz(*), b_maxz(*)
      real(kind=8)      tmp_time
      integer     tmp_li, refine, num, factor,
     *            tmp_nx0, tmp_ny0, tmp_nz0
      integer     filenum,   i, rc
      character(1) junk
      logical      sametime
      logical      double_equal
      external     double_equal
      include     'param.inc'
      !include     'largesmall.inc'

      logical     ltrace
      parameter ( ltrace = .true. )
      logical     ltrace2
      parameter ( ltrace2 = .true. )

      !
      ! If no indications to create a subgrid, then none should exist:
      !
      numbox = 0

      if (ltrace2) then
         write(*,*) 'level_bbox_read: Looking for level: ', li
         write(*,*) 'level_bbox_read:          at time: ', time
      end if

      status  = -1

      filenum = 9
      open(filenum,file='.cluster',status='old',iostat=rc)
      if(rc .ne. 0) then
        write(*,*) 'level_bbox_read: Unable to open .cluster file'
        call my_exit('Unable to open .cluster file.')
      end if

  5   read(filenum,*,end=10) tmp_time,tmp_li,num,tmp_nx0,tmp_ny0,tmp_nz0
      if (ltrace2) write(*,*) '          ',tmp_time,tmp_li, num
      sametime = double_equal(tmp_time,time)
      if (sametime .and. tmp_li .eq. li) then
         if(ltrace)write(*,*)' Our Entry:  t/li/num',tmp_time,tmp_li,num
         status = 1
         numbox = num
         do i = 1, numbox
            read(filenum,*,end=10) b_minx(i),b_miny(i),b_minz(i),
     *                             b_maxx(i),b_maxy(i),b_maxz(i)
         end do
         if(ltrace2)write(*,*)' Done reading boxes'
      else
         if (ltrace2) write(*,*) '   Not our entry',
     *           tmp_time - time,
     *           tmp_li   .eq. li
         do i = 1, num
            read(filenum,*,end=10) junk
         end do
         if (tmp_time .le. time .or. sametime) go to 5
      end if

 10   continue  

      if (status.eq.1.and.i.lt.numbox) then
         write(*,*) 'level_bbox_read: Unable to read in all the boxes'
         call my_exit('Unable to read in all the boxes')
      end if

      if (status.eq.1) then
         !
         ! Rationalize different refinement factors
         !   (assumes right now that resolution same for all dimensions)
         !
         if(ltrace2)write(*,*)' Rationalize for different coarse level?'
         if(ltrace2)write(*,*)' nx0 versus tmp_nx0 = ',nx0, tmp_nx0
         if (nx0.ne.tmp_nx0) then
            ! Make factor an integer or float?
            factor = NINT( (nx0-1.d0) / (tmp_nx0-1.d0) )
            write(*,*) 'Rationalizing bboxes w/ factor=',factor
            do i = 1, numbox
               if (ltrace)
     *         write(*,99) b_minx(i),b_maxx(i),
     *                     b_miny(i),b_maxy(i),
     *                     b_minz(i),b_maxz(i)
               b_minx(i) = (b_minx(i)-1) * factor + 1
               b_miny(i) = (b_miny(i)-1) * factor + 1
               b_minz(i) = (b_minz(i)-1) * factor + 1
               b_maxx(i) = (b_maxx(i)-1) * factor + 1
               b_maxy(i) = (b_maxy(i)-1) * factor + 1
               b_maxz(i) = (b_maxz(i)-1) * factor + 1
               if (ltrace) then
               write(*,*) 'Becomes...'
               write(*,99) b_minx(i),b_maxx(i),
     *                     b_miny(i),b_maxy(i),
     *                     b_minz(i),b_maxz(i)
               end if
            end do
         end if
      end if

      close(filenum)

      if(ltrace2)write(*,*)'level_bbox_read: Done.'

 99   format(2I4,'  ',2I4,'  ',2I4)
      return
      end      ! END: level_bbox_read
