cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_return_avail:                                                         cc
cc           Returns the size of the available memory in the arena            cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      integer function mem_return_avail( )
      implicit    none
      include    'mem.inc'

      mem_return_avail = arena - qptr

      return
      end     ! END: mem_return_avail

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_return_arena:                                                         cc
cc           Returns the size of the arena.                                   cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      integer function mem_return_arena( )
      implicit    none
      include    'mem.inc'

      mem_return_arena = arena

      return
      end     ! END: mem_return_arena

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_stat_brief:                                                           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mem_stat_brief
      implicit none
      include 'grid.inc'
      real(kind=8)   percentused

      percentused = (100.d0/arena)*qptr

      write(*,17)  '                Arena used:     ',qptr,
     *               '/',arena,'  ',percentused,'% (max: ',
     *               (100.d0/arena)*max_qptr,'%)'

 17   format(A,I9,A,I9,A,F4.1,A,F4.1,A)

      return
      end     ! END: subroutine mem_stat_brief

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_stat:                                                                 cc
cc            Print statistics about current memory usage.                    cc
cc            Outputs more information based on parameter "trace_level".      cc
cc                                                                            cc
cc            NB: outputs processor number                                    cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mem_stat
      implicit none
      include 'grid.inc'
      include 'glob.inc'     ! need trace_level variable
      include 'mpif.h'
      include 'mpi_stuff.inc'
      integer  i, sizefree
      real(kind=8)   percentused

      sizefree    = 0.d0

      !
      !  Arena:
      !
      percentused = (100.d0/arena)*qptr
      write(*,9)   myid, 'mem_stat: *****************'
      write(*,17)  myid, 'mem_stat: Arena used:     ',qptr,
     *               '/',arena,'  ',percentused,'% (max: ',
     *               (100.d0/arena)*max_qptr,'%)'
 17   format('[',I3,'] ',A,I9,A,I9,A,F4.1,A,F4.1,A)

      !
      !  Free blocks:
      !
      write(*,16)myid,'mem_stat: # of free blocks:     ',num_freeblocks,
     *              '/',maxfreeblocks,'              (max: ',
     *              max_num_freeblocks,' )'
      do i = 1, num_freeblocks
         percentused = (100.d0/arena)*size_free_blocks(i)
         sizefree    = sizefree + size_free_blocks(i)
         if (trace_level.gt.50) then
         write(*,12) myid,'mem_stat:   Block:       ', i
         write(*,11) myid,'mem_stat:       %used:       ',percentused
         write(*,10) myid,'mem_stat:       FreeBlock:   ',free_blocks(i)
         write(*,10) myid,'mem_stat:       End Block:   ',free_blocks(i)
     *                              + size_free_blocks(i) - 1
         write(*,10) myid,'mem_stat:       SizeBlock:   ',
     *                                size_free_blocks(i)
         end if
      end do
      percentused = (100.d0/arena)*sizefree
      if (num_freeblocks.gt.0) then
         write(*,11)myid,'mem_stat: %Free of blocks:       ',percentused
      end if
      percentused = (100.d0/arena)*size_blockslost
      if (num_blockslost.gt.0) then
         write(*,20)myid,'mem_stat: Blocks lost:          ',
     *         num_blockslost,
     *       '       Size: ',size_blockslost,'  ',percentused,'%'
 20      format('[',I3,'] ',A,I4,A,I9,A,F4.1,A)
      end if

      !
      !  Grid numbers:
      !
      write(*,18)  myid,'mem_stat: # of Grids:           ',numgrids,'/',
     * maxnumgrids,'              (max: ',max_numgrids,' )'
  18  format('[',I3,'] ',A,I5,A,I5,A,I5,A)
      write(*,18)myid, 'mem_stat: # of free grid#s:     ',num_freegrids,
     *    '/',maxnumgrids,'              (max: ',max_num_freegrids,' )'
      if (trace_level.gt.50) then
         do i = 1, num_freegrids
            write(*,12) myid, 'mem_stat:   grid number: ',
     *                  free_grid_numbers(i)
         end do
      end if
      write(*,9)   myid, 'mem_stat: *****************'

   9  format('[',I3,'] ',A)
  10  format('[',I3,'] ',A,I9)
  11  format('[',I3,'] ',A,F6.2)
  12  format('[',I3,'] ',A,I3)
  13  format('[',I3,'] ',A,F6.1,A,F3.0,A)
  14  format('[',I3,'] ',A,I3,A,I3,A)
  15  format('[',I3,'] ',A,I9,A,I9,A)
  16  format('[',I3,'] ',A,I3,A,I3,A,I4,A)

      return
      end     ! END: subroutine mem_stat

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_dealloc:                                                              cc
cc          Frees memory by one of the following:                             cc
cc              --- adding memory to list of free blocks                      cc
cc              --- finding already freed block which is contiguous           cc
cc              --- adding memory back onto arena if at end                   cc
cc              --- disposing of memory if no more storage of free blocks     cc
cc          Doesn't not check that memory wasn't already deallocated.         cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mem_dealloc( pointer, size )
      implicit    none
      integer     pointer, size
      include    'mem.inc'
      integer     endptr, i, smallest, endfreeptr

      logical     ltrace
      parameter ( ltrace = .false. )
      
      endptr = pointer + size - 1

      if (ltrace) then
         write(*,*) 'mem_dealloc: Freeing memory:'
         write(*,*) 'mem_dealloc:    pointer = ', pointer
         write(*,*) 'mem_dealloc:    size    = ', size
         write(*,*) 'mem_dealloc:    endptr  = ', endptr
         write(*,*) 'mem_dealloc:    qptr    = ', qptr
      end if

      !
      ! Check if contiguous with qptr:
      !    if so, then reset qptr
      !    in other words, the block to be free was last to
      !    be "alloc"ated
      !
      if (qptr .eq. (endptr+1) ) then
         qptr = pointer
         if (ltrace) then
            write(*,*) 'mem_dealloc:   Contiguous with qptr'
            write(*,*) 'mem_dealloc:   qptr = ',qptr
         end if
         if ( num_freeblocks .gt. 0 ) then
            !
            ! Work backwards looking for free blocks contiq. w/ qptr/arena:
            !
            do i = num_freeblocks, 1, -1
               endfreeptr = free_blocks(i) + size_free_blocks(i) - 1
               If (ltrace) then
                  write(*,*) 'mem_dealloc:   Checking other blocks...',i
                  write(*,*) 'mem_dealloc:   endfreeptr+1=',endfreeptr+1
               end if
               if ( qptr .eq. (endfreeptr+1) ) then
                  qptr = free_blocks(i)
                  If (ltrace) then
                     write(*,*) 'mem_dealloc:   Block contiquous w/qptr'
                     write(*,*) 'mem_dealloc:   qptr = ',qptr
                  end if
                  if ( i .eq. num_freeblocks) then
                     ! one less free block (even if num_freeblocks==1)
                     num_freeblocks = num_freeblocks - 1
                  else
                     ! replace block "i" with last block
                     free_blocks(i)      = free_blocks(num_freeblocks)
                     size_free_blocks(i) =
     *                                  size_free_blocks(num_freeblocks)
                     num_freeblocks = num_freeblocks - 1
                  end if
                  If (ltrace) write(*,*) 'mem_dealloc:num_freeblocks=',
     *                                                num_freeblocks
               end if
            end do
         end if
         if (ltrace) then
            write(*,*) 'mem_dealloc: Done.'
         end if
         return
      end if
      
      !
      ! If first block to be freed:
      !
      if (num_freeblocks .eq. 0) then
         num_freeblocks      = 1
         max_num_freeblocks  = num_freeblocks
         free_blocks(1)      = pointer
         size_free_blocks(1) = size
         if (ltrace) then
            write(*,*)'mem_dealloc:  First block to be freed.'
            write(*,*)'mem_dealloc:     num_freeblocks =',num_freeblocks
            write(*,*)'mem_dealloc:     free_blocks(1) =',free_blocks(1)
            write(*,*)'mem_dealloc: size_free_blocks(1)=',
     *                              size_free_blocks(1)
            write(*,*)'mem_dealloc: Done.'
         end if
         return
      end if

      !
      ! Check other free blocks to see if contiguous with this block:
      !
      i        = 1
      smallest = 1
  10  endfreeptr = free_blocks(i) + size_free_blocks(i) - 1
      if (free_blocks(i) .eq. (endptr+1)) then
         !
         ! New block contiguous with beginning of an old block:
         !
         free_blocks(i)      = pointer 
         size_free_blocks(i) = size_free_blocks(i) + size
         if (ltrace) then
            write(*,*)'mem_dealloc: Contiguous w/ begin. block',i
            write(*,*)'mem_dealloc:     free_blocks(i)=',free_blocks(i)
            write(*,*)'mem_dealloc:size_free_blocks(i)=',
     *                             size_free_blocks(i)
            write(*,*)'mem_dealloc: Done.'
         end if
         return
      else if ( (endfreeptr+1) .eq. pointer) then
         !
         ! New block contiguous with end of an old block:
         !
         size_free_blocks(i) = size_free_blocks(i) + size
         if (ltrace) then
            write(*,*)'mem_dealloc: Contiguous w/ end of block',i
            write(*,*)'mem_dealloc:     free_blocks(i)=',free_blocks(i)
            write(*,*)'mem_dealloc:size_free_blocks(i)=',
     *                             size_free_blocks(i)
            write(*,*)'mem_dealloc: Done.'
         end if
         return
      else
         !
         ! Not contiguous with block, look at next one.
         !
         i = i + 1
         if (i.le.num_freeblocks) then
            !
            ! Keep track of smallest sized free block (see below):
            !
            if ( size_free_blocks(smallest) .gt. size_free_blocks(i) ) 
     *           smallest = i
            go to 10
         end if
      end if

      !
      ! No blocks contiguous:
      !
      if (num_freeblocks .lt. maxfreeblocks) then
         if (ltrace) write(*,*) 'mem_dealloc: Adding block to free list'
         !
         ! Keep track of the free block:
         !
         num_freeblocks                   = num_freeblocks + 1
         if (num_freeblocks.gt.max_num_freeblocks)
     *                 max_num_freeblocks = num_freeblocks
         free_blocks(num_freeblocks)      = pointer
         size_free_blocks(num_freeblocks) = size
         if (ltrace) then
            write(*,*)'mem_dealloc: Adding new free block'
            write(*,*)'mem_dealloc:     num_freeblocks=',num_freeblocks
            write(*,*)'mem_dealloc:     free_blocks(n)=',
     *                          free_blocks(num_freeblocks)
            write(*,*)'mem_dealloc:size_free_blocks(n)=',
     *                          size_free_blocks(num_freeblocks)
         end if
      else
         !
         ! Replace the smallest free block
         !      if this block larger than the smallest
         !      else simply lose this memory block
         !
         if (size_free_blocks(smallest) .lt. size) then
            if (ltrace) write(*,*)'mem_dealloc:Replacing block',smallest
            num_blockslost             = num_blockslost + 1
            size_blockslost            =   size_blockslost
     *                                   + size_free_blocks(smallest)
            free_blocks(smallest)      = pointer
            size_free_blocks(smallest) = size
         else
            if (ltrace) write(*,*) 'mem_dealloc: Discarding block'
            num_blockslost  = num_blockslost  + 1
            size_blockslost = size_blockslost + size
         end if
      end if

      if (ltrace) then
         write(*,*)'mem_dealloc: Done.'
      end if

      return
      end     ! END: mem_dealloc

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_alloc:                                                                cc
cc           Try to use a previously freed block.                             cc
cc           If no block big enough, then free in the                         cc
cc           conventional way. If not enough memory, then quit.               cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      integer function mem_alloc( size )
      implicit    none
      integer     size
      include    'mem.inc'
      integer     i,     size_i,
     *            found, size_found, myid
      integer     proc_return_myid
      real(kind=8) percentmemuse 
      external    proc_return_myid
      logical     block_found
      logical     usefreeblocks
      parameter ( usefreeblocks = .true. )
      logical     ltrace
      parameter ( ltrace  = .false. )
      logical     ltrace2
      parameter ( ltrace2 = .false. )

      block_found = .false.
      myid        = proc_return_myid()

      if (size .lt. 0) then
         write(*,*) 'mem_alloc: Size cannot be negative',size
         call my_exit('mem_alloc: Size cannot be negative')
      end if

      if (ltrace) then
         write(*,*) 'mem_alloc: Allocating memory for block size:',size
         if (.not.usefreeblocks)
     *   write(*,*) 'mem_alloc:      Not using free blocks, only arena'
      end if

      !
      ! If free blocks exist, find smallest block that's big enough
      !
      if ( (num_freeblocks.gt.0) .and. usefreeblocks ) then
         i      = 1
  10     size_i = size_free_blocks(i)
         if (ltrace2)write(*,*) 'mem_alloc:   Looking at free block',i
         if (ltrace2)write(*,*) 'mem_alloc:            with size',size_i
         if (size_i.ge.size) then
            if (ltrace2)write(*,*) 'mem_alloc:    Block large enough...'
            if (block_found) then
               if ( size_i .le. size_found ) then
                  found      = i
                  size_found = size_i
               end if
            else
               block_found = .true.
               found       = i
               size_found  = size_i
            end if
         end if
         i = i + 1
         if (i .le. num_freeblocks) goto 10
      end if

      if (block_found) then
            !
            !  Smallest block which is big enough is grid "found":
            !
            mem_alloc      = free_blocks(found)
            if (ltrace2) then
               write(*,*) 'mem_alloc:       Found free block:'
               write(*,*) 'mem_alloc:     num block found = ',found
               write(*,*) 'mem_alloc: size of block found = ',size_found
               write(*,*) 'mem_alloc:        ptr location = ',mem_alloc
            end if
            if (size_found .eq. size) then
               !
               ! Replace just allocated block with last free block:
               !
               if (ltrace2)write(*,*) 'mem_alloc: free block eliminated'
               free_blocks(found)      = free_blocks(num_freeblocks)
               size_free_blocks(found) =size_free_blocks(num_freeblocks)
               num_freeblocks          = num_freeblocks - 1
            else
               !
               ! Decrease size of the free block:
               !
               free_blocks(found)      = free_blocks(found) + size
               size_free_blocks(found) = size_free_blocks(found) - size
               if (ltrace2) then
                  write(*,*) 'mem_alloc:   size free block decreased:'
                  write(*,*) 'mem_alloc:        free_block(found):',
     *                                               free_blocks(found)
                  write(*,*) 'mem_alloc:   size_free_block(found):',
     *                                          size_free_blocks(found)
               end if
            end if
            if (ltrace2) then
               write(*,*) 'mem_alloc:       Used  free block.'
            end if
      else
         if (ltrace2)write(*,*) 'mem_alloc: Allocating from bulk mem'
         if ((qptr+size).gt.arena) then
              percentmemuse = 100.d0*qptr /arena
              write(*,97)myid,'mem_alloc: Cannot allocate space:'
              write(*,99)myid,'mem_alloc: Size requested: ',size
              write(*,99)myid,'mem_alloc: Arena size:     ',arena
              write(*,99)myid,'mem_alloc:  Memory used:   ',qptr
              write(*,98)myid,'mem_alloc:%Memory used:   ',percentmemuse
              write(*,99)myid,'mem_alloc: myid:           ',myid
              write(*,97)myid,'mem_alloc:If possible,increase arena siz'
              write(*,97)myid,'mem_alloc: in mem.inc, and recompile.'
              call mem_stat
              write(*,99) myid, 'mem_alloc: Quitting.....'
              mem_alloc = -1
              call my_exit('Out of memory.          ')
         else
              mem_alloc = qptr
              qptr      = qptr + size
              if (qptr .gt. max_qptr) max_qptr = qptr
              if (ltrace) write(*,*) 'mem_alloc: qptr = ',qptr
         end if
      end if

      if (ltrace) then
         write(*,*) 'mem_alloc:             mem_alloc = ',mem_alloc
         write(*,*) 'mem_alloc:      mem_alloc + size = ',mem_alloc+size
      end if

 97   format('[',I3,'] ',A)
 98   format('[',I3,'] ',A,F10.3)
 99   format('[',I3,'] ',A,I15)

      return
      end     ! END: mem_alloc

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_writestate:                                                           cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mem_writestate(fname)
      implicit none
      character*(*) fname
      include 'glob.inc'
      include 'grid.inc'
      include 'output.inc'
      integer     i, len
      integer     mystringlength
      external    mystringlength
      character(128) name, tmpstr
      logical     ltrace
      parameter ( ltrace = .false. )

      if (ltrace) write(*,*) 'mem_writestate: Begin output to: ',fname

      !
      ! Stuff stored in mem.inc:
      !
      gft_rc=mygft_write_id_int(fname,'qptr',             qptr,    1)
      gft_rc=mygft_write_id_int(fname,'max_qptr',         max_qptr,1)
      gft_rc=mygft_write_id_int(fname,'max_num_freeblocks',
     *                                             max_num_freeblocks,1)
      gft_rc=mygft_write_id_int(fname,'num_freeblocks',   
     *                                             num_freeblocks,1)
      gft_rc=mygft_write_id_int(fname,'num_blockslost',   
     *                                             num_blockslost,1)
      gft_rc=mygft_write_id_int(fname,'size_blockslost',  
     *                                             size_blockslost,1)
      if (num_freeblocks.gt.0) then
      gft_rc=mygft_write_id_int(fname,'size_free_blocks',
     *                                             size_free_blocks,
     *                                             num_freeblocks)
      gft_rc=mygft_write_id_int(fname,'free_blocks',
     *                                             free_blocks,
     *                                             num_freeblocks)
      end if

      gft_rc=mygft_write_id_float(fname,'q',       q,       qptr    )

      !
      ! Stuff stored in grid.inc:
      !
      gft_rc=mygft_write_id_int(fname,'numgrids',  numgrids,      1)
      gft_rc=mygft_write_id_int(fname,'max_numgrids',     
     *                                             max_numgrids,  1)
      gft_rc=mygft_write_id_int(fname,'num_freegrids',    
     *                                             num_freegrids, 1)
      gft_rc=mygft_write_id_int(fname,'max_num_freegrids',
     *                                             max_num_freegrids,1)
      if (num_freegrids.gt.0)
     *gft_rc=mygft_write_id_int(fname,'free_grid_numbers',
     *                                             free_grid_numbers,
     *                                             num_freegrids)

      gft_rc=mygft_write_id_float(fname,'gr_h',    gr_h,    numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_t',    gr_t,    numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_work', gr_work, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_maxx', gr_maxx, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_maxy', gr_maxy, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_maxz', gr_maxz, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_minx', gr_minx, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_miny', gr_miny, numgrids)
      gft_rc=mygft_write_id_float(fname,'gr_minz', gr_minz, numgrids)

      gft_rc=mygft_write_id_int(fname,'gr_own',    gr_own,     numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_count',  gr_count,   numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_level',  gr_level,   numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_parent', gr_parent,  numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_child',  gr_child,   numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_sibling',gr_sibling, numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_size',   gr_size,    numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_nx',     gr_nx,      numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_ny',     gr_ny,      numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_nz',     gr_nz,      numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_x',      gr_x,       numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_y',      gr_y,       numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_z',      gr_z,       numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_alive',  gr_alive,   numgrids)
      gft_rc=mygft_write_id_int(fname,'gr_iter',   gr_iter,    numgrids)

      !
      ! Field pointers from fields.inc:
      !
c     gft_rc=mygft_write_id_int(fname,'gfunc_pointer', gfunc_pointer,
c    *                                    num_gfuncs*maxnumgrids)
      name = 'gfptr'
      len  = mystringlength(name)
      do i = 1, numgrids
         call int2str(i,tmpstr)
         name(len+1:) = tmpstr
         !write(*,*) name
         gft_rc = mygft_write_id_int(fname,name,gfunc_pointer(1,i),
     *                      num_gfuncs)
      end do

      !
      ! Stuff stored in glob.inc:
      !
      gft_rc=mygft_write_id_int(fname,'steps',       steps,         1 )
      gft_rc=mygft_write_id_int(fname,'Levelp',      Levelp(0),  maxlev)
      gft_rc=mygft_write_id_int(fname,'lev_count',  lev_count(0),maxlev)
      gft_rc=mygft_write_id_float(fname,'total_work',total_work,    1 )
      gft_rc=mygft_write_id_float(fname,'maxchi',    maxchi,        1 )
      gft_rc=mygft_write_id_float(fname,'maxchi_t',  maxchi_t,      1 )
      gft_rc=mygft_write_id_int(fname,'maxchi_i',    maxchi_i,      1 )
      gft_rc=mygft_write_id_int(fname,'maxchi_j',    maxchi_j,      1 )
      gft_rc=mygft_write_id_int(fname,'maxchi_k',    maxchi_k,      1 )
      gft_rc=mygft_write_id_int(fname,'maxchi_l',    maxchi_l,      1 )
      gft_rc=mygft_write_id_int(fname,'fileout',     fileout,       1 )

      if (ltrace) write(*,*) 'mem_writestate: End output to: ',fname

      return
      end     ! END: mem_writestate

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cc                                                                            cc
cc  mem_readstate:                                                            cc
cc                                                                            cc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine mem_readstate(fname)
      implicit none
      character*(*) fname
      include 'glob.inc'
      include 'grid.inc'
      include 'output.inc'
      integer     i, len
      integer     mystringlength
      external    mystringlength
      character(128) name, tmpstr
      logical     ltrace
      parameter ( ltrace = .false. )

      if (ltrace) write(*,*) 'mem_readstate: Begin read from: ',fname

      !
      ! Stuff stored in mem.inc:
      !
      !gft_rc=mygft_read_id_int(fname,'arena',            arena,1)
      gft_rc=mygft_read_id_int(fname,'qptr',             qptr,1)
         if (qptr .gt. arena) then
            write(*,*) 'mem_readstate: Problem.'
            stop
         end if
      gft_rc=mygft_read_id_int(fname,'max_qptr',         max_qptr,1)
      gft_rc=mygft_read_id_int(fname,'max_num_freeblocks',
     *                                           max_num_freeblocks,1)
      gft_rc=mygft_read_id_int(fname,'num_freeblocks',   
     *                                               num_freeblocks,1)
      gft_rc=mygft_read_id_int(fname,'num_blockslost',   
     *                                               num_blockslost,1)
      gft_rc=mygft_read_id_int(fname,'size_blockslost',  
     *                                               size_blockslost,1)

      if (num_freeblocks.gt.0) then
      gft_rc=mygft_read_id_int(fname,'size_free_blocks',
     *                                               size_free_blocks,
     *                                                   num_freeblocks)
      gft_rc=mygft_read_id_int(fname,'free_blocks',free_blocks,
     *                                               num_freeblocks)
      end if

      gft_rc=mygft_read_id_float(fname,'q',q, qptr)

      !
      ! Stuff stored in grid.inc:
      !
      gft_rc=mygft_read_id_int(fname,'numgrids',         numgrids,1)
      gft_rc=mygft_read_id_int(fname,'max_numgrids',     max_numgrids,1)
      gft_rc=mygft_read_id_int(fname,'num_freegrids',    
     *                                                num_freegrids,1)
      gft_rc=mygft_read_id_int(fname,'max_num_freegrids',
     *                                             max_num_freegrids,1)
      if (num_freegrids.gt.0)
     *gft_rc=mygft_read_id_int(fname,'free_grid_numbers',
     *                                                free_grid_numbers,
     *                                                    num_freegrids)

      gft_rc=mygft_read_id_float(fname,'gr_h',   gr_h,    numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_t',   gr_t,    numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_work',gr_work, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_maxx',gr_maxx, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_maxy',gr_maxy, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_maxz',gr_maxz, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_minx',gr_minx, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_miny',gr_miny, numgrids)
      gft_rc=mygft_read_id_float(fname,'gr_minz',gr_minz, numgrids)

      gft_rc=mygft_read_id_int(fname,'gr_own',     gr_own,     numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_count',   gr_count,   numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_level',   gr_level,   numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_parent',  gr_parent,  numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_child',   gr_child,   numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_sibling', gr_sibling, numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_size',    gr_size,    numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_nx',      gr_nx,      numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_ny',      gr_ny,      numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_nz',      gr_nz,      numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_x',       gr_x,       numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_y',       gr_y,       numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_z',       gr_z,       numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_alive',   gr_alive,   numgrids)
      gft_rc=mygft_read_id_int(fname,'gr_iter',    gr_iter,    numgrids)

      !
      ! Field pointers from fields.inc:
      !
c     gft_rc=mygft_read_id_int(fname,'gfunc_pointer', gfunc_pointer,
c    *                                    num_gfuncs*maxnumgrids)
      name = 'gfptr'
      len  = mystringlength(name)
      do i = 1, numgrids
         call int2str(i,tmpstr)
         name(len+1:) = tmpstr
         !write(*,*) i, name
      gft_rc=mygft_read_id_int(fname,name,gfunc_pointer(1,i),num_gfuncs)
      end do

      !
      ! Stuff stored in glob.inc:
      !
      gft_rc=mygft_read_id_int(fname,'steps',        steps,         1 )
      gft_rc=mygft_read_id_int(fname,'Levelp',      Levelp(0),   maxlev)
      gft_rc=mygft_read_id_int(fname,'lev_count',   lev_count(0),maxlev)
      gft_rc=mygft_read_id_float(fname,'total_work', total_work,    1 )
      gft_rc=mygft_read_id_float(fname,'maxchi',     maxchi,        1 )
      gft_rc=mygft_read_id_float(fname,'maxchi_t',   maxchi_t,      1 )
      gft_rc=mygft_read_id_int(fname,'maxchi_i',     maxchi_i,      1 )
      gft_rc=mygft_read_id_int(fname,'maxchi_j',     maxchi_j,      1 )
      gft_rc=mygft_read_id_int(fname,'maxchi_k',     maxchi_k,      1 )
      gft_rc=mygft_read_id_int(fname,'maxchi_l',     maxchi_l,      1 )
      gft_rc=mygft_read_id_int(fname,'fileout',      fileout,       1 )

      !write(*,*) 'fileout:  ',fileout
      !write(*,*) 'allowedl: ',allowedl
      if (ltrace) write(*,*) 'mem_readstate: End   read from: ',fname

      return
      end     ! END: mem_readstate

